#!/usr/bin/env python

import io
import json
import os
import re
import unicodedata as ud
from contextlib import redirect_stderr
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import List
from unittest import TestCase, main, mock

from pydantic import ValidationError

from g2p import exceptions, make_g2p
from g2p.exceptions import InvalidNormalization, NeuralDependencyError
from g2p.log import LOGGER
from g2p.mappings import Mapping, Rule
from g2p.mappings.utils import NORM_FORM_ENUM, RULE_ORDERING_ENUM, normalize
from g2p.tests.public import __file__ as public_data
from g2p.transducer import Transducer


def rules_from_strings(*mapping: str) -> List[dict]:
    """Quick pseudo constructor for unit testing of mappings"""
    rules = []
    for rule in mapping:
        key, value = rule.split(":")
        rules.append({"in": key, "out": value})
    return rules


class MappingTest(TestCase):
    """Basic Mapping Test"""

    def setUp(self):
        self.test_mapping_no_norm = Mapping(
            rules=[
                {"in": "\u00e1", "out": "\u00e1"},
                {"in": "\u0061\u0301", "out": "\u0061\u0301"},
            ],
            norm_form="none",
        )
        self.test_mapping_norm = Mapping(rules=[{"in": "\u00e1", "out": "\u00e1"}])
        with open(
            os.path.join(os.path.dirname(public_data), "git_to_ipa.json"),
            encoding="utf8",
        ) as f:
            self.json_map = json.load(f)

    def test_find_mappings(self):
        rules_mapping = make_g2p("str", "str-ipa")
        self.assertIsNone(rules_mapping.transducers[-1].mapping.type)

    def test_neural_not_available_raises_exception(self):
        from g2p.mappings.utils import (
            deep_phonemizer_handler,
            download_huggingface_model,
            has_neural_support,
        )

        # Call things that need neural with a mock context simulating not having
        # installed the dependencies, even when they are actually installed
        with mock.patch("g2p.mappings.utils.has_neural_support", return_value=False):
            with self.assertRaises(NeuralDependencyError):
                make_g2p("str", "str-ipa", neural=True)
            with self.assertRaises(NeuralDependencyError):
                deep_phonemizer_handler(Path(), {}, "foo")
            with self.assertRaises(NeuralDependencyError):
                download_huggingface_model("foo")

        # calling has_neural_support should return a Bool without raising any exception,
        # but we can't assert the value since we don't know if neural deps are installed
        self.assertTrue(isinstance(has_neural_support(), bool))

    def test_normalization(self):
        self.assertEqual(
            ud.normalize("NFD", "\u00e1"), self.test_mapping_norm.rules[0].rule_input
        )
        self.assertNotEqual(self.test_mapping_norm.rules[0].rule_input, "\u00e1")
        self.assertEqual(self.test_mapping_norm.rules[0].rule_input, "\u0061\u0301")
        self.assertEqual(self.test_mapping_no_norm.rules[0].rule_input, "\u00e1")
        self.assertEqual(self.test_mapping_no_norm.rules[0].rule_output, "\u00e1")
        self.assertEqual(self.test_mapping_no_norm.rules[1].rule_input, "\u0061\u0301")
        self.assertEqual(self.test_mapping_no_norm.rules[1].rule_output, "\u0061\u0301")

    def test_utils_normalize(self):
        """Explicitly test our custom normalize function."""
        self.assertEqual(normalize(r"\u0061", None), "a")
        self.assertEqual(normalize("\u010d", "NFD"), "\u0063\u030c")
        self.assertEqual(normalize("\u0063\u030c", "NFC"), "\u010d")
        with self.assertRaises(InvalidNormalization):
            normalize("FOOBIE", "BLETCH")

    def test_json_map(self):
        json_map = Mapping(
            rules=self.json_map["map"],
            **{k: v for k, v in self.json_map.items() if k != "map"},
        )
        self.assertEqual(len(json_map.rules), 34)
        # This is a very old version of the config, I'm not even sure these tests should be in here at all.
        # self.assertTrue(json_map.kwargs["in_metadata"]["case_insensitive"])

    def test_no_mappings_key(self):
        with self.assertRaises(ValidationError):
            Mapping.load_mapping_from_path(
                os.path.join(
                    os.path.dirname(public_data), "mappings", "no_mappings_key.yaml"
                )
            )

    def test_improperly_initialized(self) -> None:
        mapping = Mapping(rules=[Rule(rule_input="a", rule_output="b")])
        mapping.rules = [{"rule_input": "something misguided"}]  # type: ignore
        with self.assertRaises(AttributeError):
            mapping.inventory()

    def test_as_is(self):
        """
        Test deprecated config: as_is.

        This will fail when we release g2p 3.0.0, at which point we can just delete this test.
        """

        # explicitly set as_is=False
        log_output = io.StringIO()
        with redirect_stderr(log_output):
            mapping_sorted = Mapping(
                rules=[{"in": "a", "out": "b"}, {"in": "aa", "out": "c"}], as_is=False
            )
        self.assertTrue(
            mapping_sorted.rule_ordering == RULE_ORDERING_ENUM.apply_longest_first
        )
        self.assertIn(
            "deprecated",
            log_output.getvalue(),
            "it should warn that the feature is deprecated",
        )
        self.assertIn(
            "apply-longest-first",
            log_output.getvalue(),
            "it should show the equivalent rule_ordering setting",
        )

        # explicitly set as_is=True
        log_output = io.StringIO()
        with redirect_stderr(log_output):
            mapping = Mapping(
                rules=[{"in": "a", "out": "b"}, {"in": "aa", "out": "c"}], as_is=True
            )
        self.assertFalse(
            mapping.rule_ordering == RULE_ORDERING_ENUM.apply_longest_first
        )
        self.assertIn(
            "deprecated",
            log_output.getvalue(),
            "it should warn that the feature is deprecated",
        )
        self.assertIn(
            "as-written",
            log_output.getvalue(),
            "it should show the equivalent rule_ordering setting",
        )

        # test the default (rule_ordering="as-written")
        mapping_as_is = Mapping(
            rules=[{"in": "a", "out": "b"}, {"in": "aa", "out": "c"}]
        )
        self.assertFalse(
            mapping.rule_ordering == RULE_ORDERING_ENUM.apply_longest_first
        )

        # test the alternative (rule_ordering="apply-longest-first")
        transducer = Transducer(mapping_sorted)
        transducer_as_is = Transducer(mapping_as_is)
        self.assertEqual(transducer("aa").output_string, "c")
        self.assertEqual(transducer_as_is("aa").output_string, "bb")

    def test_rule_ordering(self):
        """
        Test the config option:

        rule-ordering: 'as-written' (default)

        or

        rule-ordering: 'apply-shortest-first'
        """
        rules = [{"in": "a", "out": "b"}, {"in": "aa", "out": "c"}]

        transducer_longest_first = Transducer(
            Mapping(rules=rules, rule_ordering="apply-longest-first")
        )
        self.assertEqual(transducer_longest_first("aa").output_string, "c")

        transducer_as_written = Transducer(
            Mapping(rules=rules, rule_ordering="as-written")
        )
        self.assertEqual(transducer_as_written("aa").output_string, "bb")

        transducer_default = Transducer(Mapping(rules=rules))
        self.assertEqual(transducer_default("aa").output_string, "bb")

    def test_rule_ordering_with_indices(self):
        """a{1}b{3} should be shorter than abc"""
        rules = [{"in": "a{1}b{2}", "out": "x{1}x{2}"}, {"in": "abc", "out": "y"}]
        mapping = Mapping(rules=rules, rule_ordering="apply-longest-first")
        transducer = Transducer(mapping)
        self.assertEqual(transducer("abc").output_string, "y")

    def test_rule_ordering_given_invalid_value(self):
        """
        It should log an error messages if given an invalid value for
        rule_ordering=...
        """
        rules = [{"in": "a", "out": "b"}, {"in": "aa", "out": "c"}]

        # typo in the valid setting:
        incorrect_value = "apply-longest-frist"

        log_output = io.StringIO()
        with redirect_stderr(log_output) and self.assertRaises(ValidationError):
            Mapping(rules=rules, rule_ordering=incorrect_value)

    def test_case_sensitive(self):
        mapping = Mapping(rules=[{"in": "A", "out": "b"}], case_sensitive=False)
        mapping_case_sensitive = Mapping(rules=[{"in": "A", "out": "b"}])
        transducer = Transducer(mapping)
        transducer_case_sensitive = Transducer(mapping_case_sensitive)
        self.assertEqual(transducer("a").output_string, "b")
        self.assertEqual(transducer_case_sensitive("a").output_string, "a")
        self.assertEqual(transducer("A").output_string, "b")

    def test_case_equivalencies(self):
        with self.assertRaises(exceptions.MalformedMapping):
            Mapping(rules=[{"in": "a", "out": "b"}], case_equivalencies={"a": "AA"})

    def test_escape_special(self):
        mapping = Mapping(rules=[{"in": r"\d", "out": "digit"}])
        mapping_escaped = Mapping(
            rules=[{"in": r"\d", "out": "b"}], escape_special=True
        )
        mapping_input_and_output_special_escaped = Mapping(
            rules=[{"in": "&", "out": "&"}], escape_special=True
        )
        mapping_specific_from_fpcc = Mapping(
            rules=[{"in": r"^", "out": "A"}, {"in": "o", "out": r"."}],
            rule_ordering="apply-longest-first",
            escape_special=True,
        )
        transducer = Transducer(mapping)
        transducer_escaped = Transducer(mapping_escaped)
        transducer_escaped_input_output = Transducer(
            mapping_input_and_output_special_escaped
        )
        transducer_fpcc = Transducer(mapping_specific_from_fpcc)
        self.assertEqual(transducer("1").output_string, "digit")
        self.assertEqual(transducer(r"\d").output_string, r"\d")
        self.assertEqual(transducer_escaped("1").output_string, "1")
        self.assertEqual(transducer_escaped(r"\d").output_string, "b")
        self.assertEqual(transducer_escaped_input_output("&").output_string, "&")
        self.assertEqual(transducer_fpcc("^o").output_string, "A.")

    def test_norm_form(self):
        mapping_nfc = Mapping(rules=[{"in": "a\u0301", "out": "a"}])  # Defaults to NFC
        mapping_nfd = Mapping(rules=[{"in": "a\u0301", "out": "a"}], norm_form="NFD")
        mapping_none = Mapping(rules=[{"in": "a\u0301", "out": "a"}], norm_form=False)

        transducer_nfc = Transducer(mapping_nfc)
        transducer_nfd = Transducer(mapping_nfd)
        transducer_none = Transducer(mapping_none)

        self.assertEqual(transducer_nfc("a\u0301").output_string, "a")
        self.assertEqual(transducer_nfc("\u00e1").output_string, "a")
        self.assertEqual(transducer_nfd("a\u0301").output_string, "a")
        self.assertEqual(transducer_nfd("\u00e1").output_string, "a")
        self.assertEqual(transducer_none("a\u0301").output_string, "a")
        self.assertEqual(transducer_none("\u00e1").output_string, "\u00e1")

    def test_reverse(self):
        mapping = Mapping(rules=[{"in": "a", "out": "b"}])
        mapping_reversed = Mapping(rules=[{"in": "a", "out": "b"}], reverse=True)
        transducer = Transducer(mapping)
        transducer_reversed = Transducer(mapping_reversed)
        self.assertEqual(transducer("a").output_string, "b")
        self.assertEqual(transducer("b").output_string, "b")
        self.assertEqual(transducer_reversed("a").output_string, "a")
        self.assertEqual(transducer_reversed("b").output_string, "a")

    def test_minimal(self):
        mapping = Mapping.load_mapping_from_path(
            os.path.join(
                os.path.dirname(public_data), "mappings", "minimal_config-g2p.yaml"
            )
        )
        transducer = Transducer(mapping)
        self.assertEqual(transducer("abb").output_string, "aaa")
        self.assertEqual(transducer("a").output_string, "a")
        self.assertFalse(
            mapping.rule_ordering == RULE_ORDERING_ENUM.apply_longest_first
        )
        self.assertFalse(mapping.case_sensitive)
        self.assertTrue(mapping.escape_special)
        self.assertEqual(mapping.norm_form, NORM_FORM_ENUM.NFD)
        self.assertTrue(mapping.reverse)

    def test_abbreviations(self):
        mapping = Mapping.load_mapping_from_path(
            os.path.join(
                os.path.dirname(public_data), "mappings", "abbreviation_config-g2p.yaml"
            )
        )
        self.assertEqual(mapping.rules[0].rule_input, "i|u")
        self.assertEqual(mapping.rules[1].rule_input, "a|e|i|o|u")
        transducer = Transducer(mapping)
        self.assertEqual(transducer("i").output_string, "1")
        self.assertEqual(transducer("e").output_string, "2")

    def test_rule_ordering_from_config(self):
        """
        Same as test_minimal, but uses "rule-ordering" instead of "as-is" in the config.
        """
        mapping = Mapping.load_mapping_from_path(
            os.path.join(os.path.dirname(public_data), "mappings", "rule-ordering.yaml")
        )
        transducer = Transducer(mapping)
        self.assertEqual(transducer("abb").output_string, "aaa")
        self.assertEqual(transducer("a").output_string, "a")
        self.assertTrue(mapping.rule_ordering == RULE_ORDERING_ENUM.apply_longest_first)
        self.assertEqual(mapping.rule_ordering, RULE_ORDERING_ENUM.apply_longest_first)
        self.assertFalse(mapping.case_sensitive)
        self.assertTrue(mapping.escape_special)
        self.assertEqual(mapping.norm_form, NORM_FORM_ENUM.NFD)
        self.assertTrue(mapping.reverse)

    def test_null_input(self):
        with self.assertLogs(LOGGER, "WARNING"):
            Mapping(rules=[{"in": "", "out": "a"}])

    def test_no_escape(self):
        mapping = Mapping(
            rules_path=os.path.join(
                os.path.dirname(public_data), "mappings", "no_escape.csv"
            )
        )
        transducer = Transducer(mapping)
        self.assertEqual(transducer("?").output_string, "ʔ")

    def test_invalid_regex(self):
        rules = [{"in": "fo(o", "out": "bar"}]
        with self.assertLogs(LOGGER, level="ERROR"):
            with self.assertRaises(exceptions.MalformedMapping) as cm:
                _ = Mapping(rules=rules)
        self.assertIn("regex", cm.exception.message)

    def test_invalid_rules_json(self):
        rules = [{"in": "a"}, {"out": "c"}]
        with self.assertRaises(ValidationError):
            Mapping(rules=rules)

    def test_invalid_rules_csv(self):
        tf = NamedTemporaryFile(
            prefix="test_invalid_rules_", mode="w", suffix=".csv", delete=False
        )
        tf.write("good-in,good-out\n\ngood-in-no-out\n")
        tf.close()
        with self.assertRaises(exceptions.MalformedMapping):
            Mapping(rules_path=tf.name)
        os.unlink(tf.name)

    def test_invalid_rules_filetype(self):
        tf = NamedTemporaryFile(
            prefix="test_invalid_rules_", mode="w", suffix=".foo", delete=False
        )
        tf.write("good-in,good-out\n\ngood-in-no-out\n")
        tf.close()
        with self.assertRaises(exceptions.IncorrectFileType):
            Mapping(rules_path=tf.name)
        os.unlink(tf.name)

    def test_extend_and_deduplicate(self):
        mapping1 = Mapping(rules=rules_from_strings("a:b", "c:d", "g:h"))
        mapping2 = Mapping(rules=rules_from_strings("a:x", "c:d", "e:f"))
        extend_ref = Mapping(
            rules=rules_from_strings("a:b", "c:d", "g:h", "a:x", "c:d", "e:f")
        )
        mapping1.extend(mapping2)
        self.assertEqual(mapping1.rules, extend_ref.rules)
        dedup_ref = Mapping(rules=rules_from_strings("a:b", "c:d", "g:h", "a:x", "e:f"))
        mapping1.deduplicate()
        self.assertEqual(mapping1.rules, dedup_ref.rules)

    def test_g2p_studio_csv(self):
        # Ensure that a single CSV file from Studio works properly
        with self.assertLogs(LOGGER, level="WARNING"):  # silence "" input warnings
            mapping = Mapping(
                rules_path=os.path.join(
                    os.path.dirname(public_data), "mappings", "g2p_studio.csv"
                )
            )
        transducer = Transducer(mapping)
        self.assertEqual(
            transducer("Jouni haluaa juoda kahvia").output_string,
            "Jouni hɑluɑː juodɑ kɑhviɑ",
        )
        # Concatenate them (this is not a good idea) and make sure it works anyway
        tf = NamedTemporaryFile(
            prefix="test_g2p_g2p_",
            mode="w",
            suffix=".csv",
            delete=False,
            encoding="utf8",
        )
        with open(
            os.path.join(os.path.dirname(public_data), "mappings", "g2p_studio.csv"),
            encoding="utf8",
        ) as fh:
            tf.write(fh.read())
        # In fact you can't concatenate them anyway. They don't end in newline.
        tf.write("\n")
        with open(
            os.path.join(os.path.dirname(public_data), "mappings", "g2p_studio2.csv"),
            encoding="utf8",
        ) as fh:
            tf.write(fh.read())
        tf.close()
        with self.assertLogs(LOGGER, level="WARNING"):  # silence "" input warnings
            mapping = Mapping(rules_path=tf.name)
        transducer = Transducer(mapping)
        self.assertEqual(
            transducer("tee on herkullista").output_string, "teː on herkullistɑ"
        )
        os.unlink(tf.name)

    def test_no_reprocess(self):
        """Ensure that attempting to reprocess a mapping is an error."""
        with self.assertRaises(AssertionError):
            self.test_mapping_norm.process_model_specs()
        with self.assertRaises(ValidationError):
            _ = Mapping(
                rules=[{"in": "a", "out": "b", "match_pattern": re.compile("XOR OTA")}]
            )
        with self.assertRaises(ValidationError):
            _ = Mapping(
                rules=[
                    {
                        "in": "a",
                        "out": "b",
                        "intermediate_form": re.compile("HACKEM MUCHE"),
                    }
                ]
            )


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import os
from unittest import TestCase, main, mock

from g2p import make_g2p
from g2p.exceptions import MalformedMapping, NeuralDependencyError
from g2p.mappings import Mapping
from g2p.tests.public import PUBLIC_DIR
from g2p.transducer import CompositeTransducer, Transducer, normalize_edges

# Disable the attr-defined mypy error so that we can use setUpClass sanely
# mypy: disable-error-code="attr-defined"


class TransducerTest(TestCase):
    """Basic Transducer Test"""

    @classmethod
    def setUpClass(cls):
        cls.test_mapping_moh = Mapping.find_mapping(
            in_lang="moh-equiv", out_lang="moh-ipa"
        )
        cls.test_mapping = Mapping(
            rules=[{"in": "a", "out": "b"}], in_lang="spam", out_lang="eggs"
        )
        cls.test_mapping_rev = Mapping(
            rules=[{"in": "a", "out": "b"}],
            reverse=True,
            in_lang="eggs",
            out_lang="parrot",
        )
        cls.test_mapping_ordered_feed = Mapping(
            rules=[{"in": "a", "out": "b"}, {"in": "b", "out": "c"}]
        )
        cls.test_mapping_ordered_counter_feed = Mapping(
            rules=[{"in": "b", "out": "c"}, {"in": "a", "out": "b"}]
        )
        cls.test_longest_first = Mapping(
            rules=[{"in": "j", "out": "ʣ"}, {"in": "'y", "out": "jˀ"}]
        )
        cls.test_rules_as_written_mapping = Mapping(
            rules=[{"in": "j", "out": "ʣ"}, {"in": "'y", "out": "jˀ"}],
            rule_ordering="apply-longest-first",
        )
        cls.test_case_sensitive_mapping = Mapping(
            rules=[{"in": "'n", "out": "n̓"}], case_sensitive=True
        )
        cls.test_case_insensitive_mapping = Mapping(
            rules=[{"in": "'n", "out": "n̓"}], case_sensitive=False
        )
        cls.test_case_sensitive_transducer = Transducer(cls.test_case_sensitive_mapping)
        cls.test_case_insensitive_transducer = Transducer(
            cls.test_case_insensitive_mapping
        )
        cls.test_trans_as_written = Transducer(cls.test_longest_first)
        cls.test_trans_longest_first = Transducer(cls.test_rules_as_written_mapping)
        cls.test_trans = Transducer(cls.test_mapping)
        cls.test_trans_ordered_feed = Transducer(cls.test_mapping_ordered_feed)
        cls.test_trans_ordered_counter_feed = Transducer(
            cls.test_mapping_ordered_counter_feed
        )
        cls.test_trans_rev = Transducer(cls.test_mapping_rev)
        cls.test_trans_moh = Transducer(cls.test_mapping_moh)
        cls.test_trans_composite = CompositeTransducer(
            [cls.test_trans, cls.test_trans_rev]
        )
        cls.test_trans_composite_2 = CompositeTransducer(
            [cls.test_trans_rev, cls.test_trans]
        )
        cls.test_regex_set_transducer_sanity = Transducer(
            Mapping(rules=[{"in": "a", "out": "b", "context_before": "c"}])
        )
        cls.test_regex_set_transducer = Transducer(
            Mapping(rules=[{"in": "a", "out": "b", "context_before": "[cd]|[fgh]"}])
        )
        cls.test_deletion_transducer = Transducer(
            Mapping(rules=[{"in": "a", "out": ""}])
        )
        csv_deletion_mapping = Mapping.load_mapping_from_path(
            os.path.join(PUBLIC_DIR, "mappings", "deletion_config_csv.yaml")
        )
        cls.test_deletion_transducer_csv = Transducer(csv_deletion_mapping)
        cls.test_deletion_transducer_json = Transducer(
            Mapping.load_mapping_from_path(
                os.path.join(PUBLIC_DIR, "mappings", "deletion_config_json.yaml")
            )
        )

    def test_no_neural_dependencies(self):
        """This tests what happens if a user tries to create a neural g2p without installing the dependencies. Other neural tests (for when deps are installed are in test_neural.py module.)"""
        with mock.patch("g2p.mappings.utils.has_neural_support", return_value=False):
            with self.assertRaises(NeuralDependencyError):
                make_g2p("foo", "bar", neural=True)
            with self.assertRaises(NeuralDependencyError):
                make_g2p("str", "str-ipa", neural=True)

    def test_properties(self):
        """Test all the basic properties of transducers."""
        self.assertEqual("spam", self.test_trans.in_lang)
        self.assertEqual("eggs", self.test_trans.out_lang)
        self.assertEqual([self.test_trans], self.test_trans.transducers)
        self.assertEqual(
            [self.test_trans, self.test_trans_rev],
            self.test_trans_composite.transducers,
        )
        self.assertEqual("spam", self.test_trans_composite.in_lang)
        self.assertEqual("parrot", self.test_trans_composite.out_lang)

    def test_graph_properties(self):
        """Test all the basic properties of graphs."""
        tg = self.test_trans("abab")
        self.assertEqual("abab", tg.input_string)
        self.assertEqual("bbbb", tg.output_string)
        self.assertEqual(1, len(tg.tiers))
        self.assertEqual([(0, "a"), (1, "b"), (2, "a"), (3, "b")], tg.input_nodes)
        self.assertEqual([(0, "b"), (1, "b"), (2, "b"), (3, "b")], tg.output_nodes)
        self.assertEqual([(0, 0), (1, 1), (2, 2), (3, 3)], tg.edges)
        self.assertEqual(
            [("a", "b"), ("b", "b"), ("a", "b"), ("b", "b")], tg.pretty_edges()
        )
        self.assertEqual(1, len(tg.debugger))
        self.assertEqual(2, len(tg.debugger[0]))
        tg.input_string = "bbbb"
        self.assertEqual([(0, "b"), (1, "b"), (2, "b"), (3, "b")], tg.input_nodes)
        tg.output_string = "baba"
        self.assertEqual([(0, "b"), (1, "a"), (2, "b"), (3, "a")], tg.output_nodes)
        tg.edges = [(0, 1), (1, 0), (2, 3), (3, 2)]
        self.assertEqual([(0, 1), (1, 0), (2, 3), (3, 2)], tg.edges)
        tg.debugger = [["spam", "spam", "spam", "spam"]]
        self.assertEqual(1, len(tg.debugger))
        self.assertEqual(4, len(tg.debugger[0]))
        with self.assertRaises(ValueError):
            tg.input_nodes = ("foo", "bar", "baz")
        with self.assertRaises(ValueError):
            tg.output_nodes = ("foo", "bar", "baz")
        with self.assertRaises(ValueError):
            tg.tiers = ["spam", "spam", "eggs", "spam"]
        tg = self.test_trans("abab")
        tg += tg
        self.assertEqual("abababab", tg.input_string)
        self.assertEqual("bbbbbbbb", tg.output_string)

    def test_composite_graph_properties(self):
        """Test all the basic properties of composite graphs."""
        ctg = self.test_trans_composite("aba")
        self.assertEqual("aba", ctg.input_string)
        self.assertEqual("aaa", ctg.output_string)
        self.assertEqual(2, len(ctg.tiers))
        self.assertEqual([(0, "a"), (1, "b"), (2, "a")], ctg.input_nodes)
        self.assertEqual([(0, "a"), (1, "a"), (2, "a")], ctg.output_nodes)
        self.assertEqual(
            [[(0, 0), (1, 1), (2, 2)], [(0, 0), (1, 1), (2, 2)]], ctg.edges
        )
        self.assertEqual(
            [
                [("a", "b"), ("b", "b"), ("a", "b")],
                [("b", "a"), ("b", "a"), ("b", "a")],
            ],
            ctg.pretty_edges(),
        )
        self.assertEqual(len(ctg.tiers), len(ctg.debugger))
        ctg.input_string = "bbbb"
        self.assertEqual([(0, "b"), (1, "b"), (2, "b"), (3, "b")], ctg.input_nodes)
        ctg.output_string = "baba"
        self.assertEqual([(0, "b"), (1, "a"), (2, "b"), (3, "a")], ctg.output_nodes)
        with self.assertRaises(ValueError):
            ctg.debugger = [["spam", "spam", "spam", "spam"]]
        with self.assertRaises(ValueError):
            ctg.edges = [(0, 1), (1, 0), (2, 3), (3, 2)]
        with self.assertRaises(ValueError):
            ctg.input_nodes = ("foo", "bar", "baz")
        with self.assertRaises(ValueError):
            ctg.output_nodes = ("foo", "bar", "baz")
        with self.assertRaises(ValueError):
            ctg.tiers = ["spam", "spam", "eggs", "spam"]
        ctg = self.test_trans_composite("aba")
        ctg += ctg
        self.assertEqual("abaaba", ctg.input_string)
        self.assertEqual("aaaaaa", ctg.output_string)

    def test_ordered(self):
        transducer_feed = self.test_trans_ordered_feed("a")
        transducer_counter_feed = self.test_trans_ordered_counter_feed("a")
        # These should feed b -> c
        self.assertEqual(transducer_feed.output_string, "c")
        # These should counter-feed b -> c
        self.assertEqual(transducer_counter_feed.output_string, "b")

    def test_forward(self):
        self.assertEqual(self.test_trans("a").output_string, "b")
        self.assertEqual(self.test_trans("b").output_string, "b")

    def test_backward(self):
        self.assertEqual(self.test_trans_rev("b").output_string, "a")
        self.assertEqual(self.test_trans_rev("a").output_string, "a")

    def test_lang_import(self):
        self.assertEqual(self.test_trans_moh("kawenón:nis").output_string, "ɡɑwenṹːnis")

    def test_composite(self):
        self.assertEqual(self.test_trans_composite("aba").output_string, "aaa")
        self.assertEqual(self.test_trans_composite_2("aba").output_string, "bbb")

    def test_rule_ordering(self):
        self.assertEqual(self.test_trans_as_written("'y").output_string, "jˀ")
        self.assertEqual(self.test_trans_longest_first("'y").output_string, "ʣˀ")

    def test_case_sensitive(self):
        self.assertEqual(self.test_case_sensitive_transducer("'N").output_string, "'N")
        self.assertEqual(self.test_case_sensitive_transducer("'n").output_string, "n̓")
        self.assertEqual(self.test_case_insensitive_transducer("'N").output_string, "n̓")
        self.assertEqual(self.test_case_insensitive_transducer("'n").output_string, "n̓")

    def test_regex_set(self):
        # https://github.com/NRC-ILT/g2p/issues/15
        self.assertEqual(
            self.test_regex_set_transducer_sanity("ca").output_string, "cb"
        )
        self.assertEqual(self.test_regex_set_transducer("ca").output_string, "cb")
        self.assertEqual(self.test_regex_set_transducer("fa").output_string, "fb")

    def test_deletion(self):
        tg = self.test_deletion_transducer("a")
        self.assertEqual(tg.output_string, "")
        self.assertEqual(tg.pretty_edges(), [("a", None)])
        self.assertEqual(self.test_deletion_transducer_csv("a").output_string, "")
        self.assertEqual(self.test_deletion_transducer_json("a").output_string, "")

    def test_case_preservation(self):
        mapping = Mapping(
            rules=[
                {"in": "'a", "out": "b"},
                {"in": "e\u0301", "out": "f"},
                {"in": "tl", "out": "λ"},
            ],
            case_sensitive=False,
            preserve_case=True,
            norm_form="NFC",
            case_equivalencies={"λ": "\u2144"},
        )
        transducer = Transducer(mapping)
        self.assertEqual(transducer("'a").output_string, "b")
        self.assertEqual(transducer("'A").output_string, "B")
        self.assertEqual(transducer("E\u0301").output_string, "F")
        self.assertEqual(transducer("e\u0301").output_string, "f")
        # Test what happens in Heiltsuk. \u03BB (λ) should be capitalized as \u2144 (⅄)
        self.assertEqual(transducer("TLaba").output_string, "\u2144aba")
        self.assertEqual(transducer("tlaba").output_string, "λaba")
        # I guess it's arguable what should happen here, but I'll just change case if any of the characters are differently cased
        self.assertEqual(transducer("Tlaba").output_string, "\u2144aba")
        # case equivalencies that are not the same length cause indexing errors in the current implementation
        with self.assertRaises(MalformedMapping):
            Mapping(
                rules=[
                    {"in": "'a", "out": "b"},
                    {"in": "e\u0301", "out": "f"},
                    {"in": "tl", "out": "λ"},
                ],
                case_sensitive=False,
                preserve_case=True,
                norm_form="NFC",
                case_equivalencies={"λ": "\u2144\u2144\u2144"},
            )

        with self.assertRaises(MalformedMapping):
            _ = Mapping(
                rules=[{"in": "a", "out": "b"}],
                case_sensitive=True,
                preserve_case=True,
            )

    def test_normalize_edges(self):
        # Remove non-deletion edges with the same index as deletions
        bad_edges = [
            (0, 0),
            (1, None),
            (1, 1),
            (2, 2),
            (3, None),
            (3, 1),
            (3, 2),
            (4, 4),
        ]
        self.assertEqual(
            normalize_edges(bad_edges), [(0, 0), (1, 0), (2, 2), (3, 2), (4, 4)]
        )
        # Sort edges on inputs and suppress duplicates
        bad_edges = [(4, 0), (1, 3), (1, 2), (2, 5)]
        self.assertEqual(normalize_edges(bad_edges), [(1, 3), (1, 2), (2, 5), (4, 0)])
        bad_edges = [(4, 0), (1, 3), (1, 3), (1, 2), (2, 5)]
        self.assertEqual(normalize_edges(bad_edges), [(1, 3), (1, 2), (2, 5), (4, 0)])
        # Map None to previous if it exists
        bad_edges = [(0, 0), (1, None), (2, 1)]
        self.assertEqual(normalize_edges(bad_edges), [(0, 0), (1, 0), (2, 1)])
        bad_edges = [(0, 0), (1, None), (2, None), (3, None)]
        self.assertEqual(normalize_edges(bad_edges), [(0, 0), (1, 0), (2, 0), (3, 0)])
        bad_edges = [(0, 0), (1, None), (2, None), (3, 1), (4, None), (5, 2)]
        self.assertEqual(
            normalize_edges(bad_edges), [(0, 0), (1, 0), (2, 0), (3, 1), (4, 1), (5, 2)]
        )
        # Map None to next if it exists
        bad_edges = [(0, None), (2, 1)]
        self.assertEqual(normalize_edges(bad_edges), [(0, 1), (2, 1)])
        bad_edges = [(0, None), (1, None), (2, 1)]
        self.assertEqual(normalize_edges(bad_edges), [(0, 1), (1, 1), (2, 1)])
        # Otherwise leave it as None
        bad_edges = []
        self.assertEqual(normalize_edges(bad_edges), bad_edges)
        bad_edges = [(0, None)]
        self.assertEqual(normalize_edges(bad_edges), bad_edges)
        bad_edges = [(0, None), (1, None)]
        self.assertEqual(normalize_edges(bad_edges), bad_edges)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import json
import os
import re
import shutil
import tempfile
from contextlib import contextmanager
from pathlib import Path
from unittest import SkipTest, TestCase, main, mock

import jsonschema
import pydantic
import yaml
from click.testing import CliRunner

import g2p._version
from g2p.cli import (
    convert,
    doctor,
    generate_mapping,
    scan,
    show_mappings,
    update,
    update_schema,
)
from g2p.exceptions import NeuralDependencyError
from g2p.log import LOGGER
from g2p.mappings import MappingConfig
from g2p.mappings.langs import (
    LANGS_DIR,
    LANGS_FILE_NAME,
    NETWORK_FILE_NAME,
    load_langs,
    load_network,
)
from g2p.tests.public.data import DATA_DIR, load_public_test_data


def set_g2p_version(version_tuple, version_string=None):
    if version_string is None:
        version_string = ".".join(str(part) for part in version_tuple)
    g2p._version.VERSION = g2p._version.__version__ = g2p._version.version = (
        version_string
    )
    g2p._version.__version_tuple__ = g2p._version.version_tuple = tuple(version_tuple)


def relaxed_int(int_or_str) -> int:
    """Parse a version component returning only its numerical prefix.

    Motivation: in CI, sometimes we end up with a version like 2.3.dev0
    E.g.: 42 -> 42, 1dev2 -> 1, dev -> 0
    """
    if isinstance(int_or_str, int):
        return int_or_str
    m = re.search(r"^[0-9]+", int_or_str)
    return int(m.group()) if m else 0


@contextmanager
def monkey_patch_g2p_version(increment_tuple):
    saved_version = g2p._version.VERSION
    saved_version_tuple = g2p._version.version_tuple
    incremented_version = list(g2p._version.version_tuple)
    while len(incremented_version) < len(increment_tuple):
        incremented_version.append(0)
    for part, increment in enumerate(increment_tuple):
        incremented_version[part] = relaxed_int(incremented_version[part]) + increment
    set_g2p_version(incremented_version)
    yield
    set_g2p_version(saved_version_tuple, saved_version)


class CliTest(TestCase):
    """Test suite for the g2p Command Line Interface"""

    def setUp(self):
        self.runner = CliRunner()

    def test_update(self):
        result = self.runner.invoke(update)
        # Test running in another directory
        with tempfile.TemporaryDirectory() as tmpdir:
            lang1_dir = os.path.join(tmpdir, "lang1")
            os.mkdir(lang1_dir)
            mappings_dir = os.path.join(DATA_DIR, "..", "mappings")
            for name in os.listdir(mappings_dir):
                if name.startswith("minimal."):
                    shutil.copy(
                        os.path.join(mappings_dir, name), os.path.join(lang1_dir, name)
                    )
            shutil.copy(
                os.path.join(mappings_dir, "minimal_configs.yaml"),
                os.path.join(lang1_dir, "config-g2p.yaml"),
            )
            result = self.runner.invoke(update, ["-i", tmpdir])
            langs_json = os.path.join(tmpdir, LANGS_FILE_NAME)
            network_pkl = os.path.join(tmpdir, NETWORK_FILE_NAME)
            self.assertTrue(os.path.exists(langs_json))
            self.assertTrue(os.path.exists(network_pkl))

        # Make sure it produces output
        with tempfile.TemporaryDirectory() as tmpdir:
            result = self.runner.invoke(update, ["-o", tmpdir])
            self.assertEqual(result.exit_code, 0)
            langs_json = os.path.join(tmpdir, LANGS_FILE_NAME)
            network_pkl = os.path.join(tmpdir, NETWORK_FILE_NAME)
            self.assertTrue(os.path.exists(langs_json))
            self.assertTrue(os.path.exists(network_pkl))
            langs = load_langs(langs_json)
            self.assertTrue(langs is not None)
            network = load_network(network_pkl)
            self.assertTrue(network is not None)
            # Corrupt the output and make sure we still can run
            with open(langs_json, "wb") as fh:
                fh.write(b"spam spam spam")
            with open(network_pkl, "wb") as fh:
                fh.write(b"eggs bacon spam")
            with self.assertLogs(LOGGER, "WARNING"):
                langs = load_langs(langs_json)
            self.assertTrue(langs is not None)
            with self.assertLogs(LOGGER, "WARNING"):
                network = load_network(network_pkl)
            self.assertTrue(network is not None)
        # Make sure it fails meaningfully on invalid input
        with tempfile.TemporaryDirectory() as tmpdir:
            bad_langs_dir = os.path.join(DATA_DIR, "..", "mappings", "bad_langs")
            result = self.runner.invoke(update, ["-i", bad_langs_dir, "-o", tmpdir])
            self.assertNotEqual(result.exit_code, 0)
            self.assertIn("mappings", str(result.exception))
        with tempfile.TemporaryDirectory() as tmpdir:
            bad_langs_dir = os.path.join(DATA_DIR, "..", "mappings", "bad_langs2")
            result = self.runner.invoke(update, ["-i", bad_langs_dir, "-o", tmpdir])
            self.assertEqual(result.exit_code, 0)

    def test_wont_update_schema_with_pydantic_ge29(self):
        """Make sure schema update does nothing when Pydantic>=2.9"""
        pydantic_major, pydantic_minor, _ = pydantic.VERSION.split(".", 3)
        if (int(pydantic_major), int(pydantic_minor)) < (2, 9):
            raise SkipTest("This test is only meaningful with pydantic>=2.9")

        result = self.runner.invoke(update_schema)
        self.assertEqual(result.exit_code, 0)
        self.assertIn("Please use Pydantic", result.output)

    def test_update_schema_with_pydantic_lt29(self):
        """Make sure schema update works (requires Pydantic<2.9)"""
        # Skip this test if the currently installed pydantic version is >= 2.9
        pydantic_major, pydantic_minor, _ = pydantic.VERSION.split(".", 3)
        if (int(pydantic_major), int(pydantic_minor)) >= (2, 9):
            raise SkipTest("This test is only meaningful with pydantic<2.9")

        # It's an error for the currently saved schema to be out of date
        result = self.runner.invoke(update_schema)
        self.assertEqual(result.exit_code, 0)
        self.assertIn("up to date", result.output)

        with tempfile.TemporaryDirectory() as tmpdir:
            # Exercise writing a new schema to disk even if up to date
            result = self.runner.invoke(update_schema, ["-o", tmpdir])
            self.assertEqual(result.exit_code, 0)
            self.assertIn("Wrote", result.output)

            # Reload the written schema for further unit tests
            (major, minor, *_rest) = g2p._version.version_tuple
            major_minor = f"{major}.{minor}"
            with open(
                Path(tmpdir) / f"g2p-config-schema-{major_minor}.json",
                encoding="utf8",
            ) as f:
                schema = json.load(f)

            # A second run will necessarily already be up to date even if the patch is bumped
            with monkey_patch_g2p_version((0, 0, +1)):
                result_rerun = self.runner.invoke(update_schema, ["-o", tmpdir])
                self.assertEqual(result_rerun.exit_code, 0)
                self.assertIn("already up to date", result_rerun.output)

            # Monkey patch the version to test a previous version still being up to date
            with monkey_patch_g2p_version((+0, +1)):
                result_new = self.runner.invoke(update_schema, ["-o", tmpdir])
                self.assertEqual(result_new.exit_code, 0)
                self.assertIn("still up to date", result_new.output)

            # Monkey patch the version and the model to require a schema update
            with monkey_patch_g2p_version((+1, +0)):
                saved_doc = MappingConfig.__doc__
                MappingConfig.__doc__ = "Changed docstring"
                result_update = self.runner.invoke(update_schema, ["-o", tmpdir])
                MappingConfig.__doc__ = saved_doc
                self.assertEqual(result_update.exit_code, 0)
                self.assertIn("Wrote", result_update.output)

            # Require a schema update when it's already written: that's an error
            with monkey_patch_g2p_version((+1, +0)):
                result_bad_update = self.runner.invoke(update_schema, ["-o", tmpdir])
                self.assertNotEqual(result_bad_update.exit_code, 0)
                self.assertIn("but is not up to date", result_bad_update.output)

        # Validate all configurations against the current schema, quietly unless there's an error:
        for config in Path(LANGS_DIR).glob("**/config-g2p.yaml"):
            with open(config, encoding="utf8") as f:
                config_yaml = yaml.safe_load(f)
            try:
                jsonschema.validate(config_yaml, schema=schema)
            except jsonschema.exceptions.ValidationError:
                self.fail(f"Error validating {config}")

    def test_convert(self):
        """Running all g2p convert test cases found in g2p/tests/public/data"""
        langs_to_test = load_public_test_data()
        error_count = 0
        for tok_option in [["--tok", "--check"], ["--no-tok"]]:
            for (
                in_lang,
                out_lang,
                word_to_convert,
                reference_string,
                *_,
                fileline,
            ) in langs_to_test:
                result = self.runner.invoke(
                    convert, [*tok_option, word_to_convert, in_lang, out_lang]
                )
                self.assertEqual(result.exit_code, 0)
                if "--no-tok" not in tok_option:
                    output_string = result.stdout.strip()
                    if reference_string.strip() not in output_string:
                        LOGGER.warning(
                            f"test_cli.py for {fileline}: {in_lang}->{out_lang} mapping error: '{word_to_convert}' "
                            f"should map to '{reference_string}', got '{output_string}' (with {tok_option})."
                        )
                        if error_count == 0:
                            first_failed_test = (
                                in_lang,
                                out_lang,
                                word_to_convert,
                                tok_option,
                                reference_string,
                            )
                        error_count += 1

        if error_count > 0:
            (
                in_lang,
                out_lang,
                word_to_convert,
                tok_option,
                reference_string,
            ) = first_failed_test
            output_string = self.runner.invoke(
                convert,
                [*tok_option, word_to_convert, in_lang, out_lang],
            ).stdout.strip()
            self.assertEqual(
                output_string,
                reference_string.strip(),
                f"{in_lang}->{out_lang} mapping error for '{word_to_convert}'.\n"
                "Look for warnings in the log for any more mapping errors",
            )

    def test_convert_neural(self):
        with mock.patch("g2p.mappings.utils.has_neural_support", return_value=False):
            with self.assertRaises(NeuralDependencyError):
                result = self.runner.invoke(
                    convert, ["--neural", "hello world", "str", "str-ipa"]
                )
                raise result.exception  # type: ignore[misc]

    def test_doctor(self):
        result = self.runner.invoke(doctor, "-m fra")
        self.assertEqual(result.exit_code, 2)

        result = self.runner.invoke(doctor, "-m fra-ipa")
        self.assertEqual(result.exit_code, 0)

        # Disable this test: it's very slow (8s, just by itself) and does not assert
        # anything useful.
        # Migrated to test_doctor_expensive.py so we can still run it, manually or via
        # ./run.py all.
        # result = self.runner.invoke(doctor)
        # self.assertEqual(result.exit_code, 0)
        # self.assertGreaterEqual(len(result.stdout), 10000)

        result = self.runner.invoke(doctor, "-m eng-arpabet")
        self.assertEqual(result.exit_code, 0)
        self.assertIn("No checks implemented", result.stdout)

    def test_doctor_lists(self):
        result = self.runner.invoke(doctor, "--list-all")
        self.assertEqual(result.exit_code, 0)
        self.assertIn("eng-arpabet:", result.stdout)
        self.assertIn("eng-ipa:", result.stdout)

        result = self.runner.invoke(doctor, "--list-ipa")
        self.assertEqual(result.exit_code, 0)
        self.assertNotIn("eng-arpabet:", result.stdout)
        self.assertIn("eng-ipa:", result.stdout)

    def test_scan_fra(self):
        """Test g2p scan with all French characters, in NFC and NFD"""
        for paragram_file in ["fra_panagrams.txt", "fra_panagrams_NFD.txt"]:
            result = self.runner.invoke(
                scan, ["fra", os.path.join(DATA_DIR, paragram_file)]
            )
            self.assertEqual(result.exit_code, 0)
            self.assertLogs(level="WARNING")
            diacritics = "àâéèêëîïôùûüç"
            for d in diacritics:
                self.assertNotIn(d, result.stdout)
            unmapped_chars = ":/,'-()2"
            for c in unmapped_chars:
                self.assertIn(c, result.stdout)

    def test_scan_fra_simple(self):
        # For now, unit test g2p scan using a simpler piece of French
        result = self.runner.invoke(
            scan, ["fra", os.path.join(DATA_DIR, "fra_simple.txt")]
        )
        self.assertEqual(result.exit_code, 0)
        self.assertLogs(level="WARNING")
        diacritics = "àâéèêëîïôùûüç"
        for d in diacritics:
            self.assertNotIn(d, result.stdout)
        unmapped_chars = ":,"
        for c in unmapped_chars:
            self.assertIn(c, result.stdout)

    def test_scan_str_case(self) -> None:
        result = self.runner.invoke(
            scan, ["str", os.path.join(DATA_DIR, "str_un_human_rights.txt")]
        )
        returned_set = re.search("{(.*)}", result.stdout).group(1)  # type: ignore
        self.assertEqual(result.exit_code, 0)
        self.assertLogs(level="WARNING")
        unmapped_upper = "FGR"
        for u in unmapped_upper:
            self.assertIn(u, returned_set)
        unmapped_lower = "abcdefghijklqrtwxyz"
        for low in unmapped_lower:
            self.assertIn(low, returned_set)
        mapped_upper = "ABCDEHIJKLMNOPQSTUVWXYZ"
        for u in mapped_upper:
            self.assertNotIn(u, returned_set)
        mapped_lower = "s"
        self.assertNotIn(mapped_lower, returned_set)

    def test_scan_err(self):
        results = self.runner.invoke(
            scan, ["bad_lang", os.path.join(DATA_DIR, "fra_simple.txt")]
        )
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("is not a valid value for 'LANG'", results.output)

    def test_convert_option_a(self):
        result = self.runner.invoke(convert, "-a hello eng eng-arpabet")
        self.assertIn(
            "[('h', 'HH '), ('e', 'AH '), ('ll', 'L '), ('o', 'OW ')]", result.stdout
        )

    def test_convert_option_e(self):
        result = self.runner.invoke(convert, "-e est fra eng-arpabet")
        for s in [
            "[('e', 'ɛ'), ('s', 'ɛ'), ('t', 'ɛ')]",
            "[('ɛ', 'ɛ')]",
            "[('ɛ', 'E'), ('ɛ', 'H'), ('ɛ', ' ')]",
        ]:
            self.assertIn(s, result.stdout)

    def test_convert_option_d(self):
        result = self.runner.invoke(convert, "-d est fra eng-arpabet")
        for s in ["'input': 'est'", "'output': 'ɛ'", "'input': 'ɛ'", "'output': 'EH '"]:
            self.assertIn(s, result.stdout)

    def test_convert_option_t(self):
        result = self.runner.invoke(convert, "-t e\\'i oji oji-ipa")
        self.assertIn("eːʔi", result.stdout)

    def test_convert_option_tl(self):
        result = self.runner.invoke(convert, "--tok-lang fra e\\'i oji oji-ipa")
        self.assertIn("eː'i", result.stdout)

    def test_generate_mapping_config(self):
        """Ensure that generate-mapping creates valid configuration."""
        # The underlying create_mapping() function is tested in
        # test_create_mapping.py, and align_to_dummy_fallback() in
        # test_fallback.py, with less expensive inputs than our real
        # g2p mappings, and with predictable results.  However, we do
        # need to ensure that it creates/updates a correct
        # configuration, so we test that here.
        with tempfile.TemporaryDirectory() as tmpdir:
            results = self.runner.invoke(
                generate_mapping, ["--ipa", "atj", "--out-dir", tmpdir]
            )
            self.assertEqual(results.exit_code, 0)
            rulespath = os.path.join(tmpdir, "atj-ipa_to_eng-ipa.json")
            self.assertTrue(os.path.exists(rulespath))
            confpath = os.path.join(tmpdir, "config-g2p.yaml")
            config = MappingConfig.load_mapping_config_from_path(confpath)
            self.assertEqual(len(config.mappings), 1)
            self.assertEqual(config.mappings[0].rules_path, Path(rulespath))
            # Run it again, should get the same result
            results = self.runner.invoke(
                generate_mapping, ["--ipa", "atj", "--out-dir", tmpdir]
            )
            self.assertEqual(results.exit_code, 0)
            config = MappingConfig.load_mapping_config_from_path(confpath)
            self.assertEqual(len(config.mappings), 1)
            self.assertEqual(config.mappings[0].rules_path, Path(rulespath))
            # Run it with a different language, should get more config
            results = self.runner.invoke(
                generate_mapping, ["--ipa", "alq", "--out-dir", tmpdir]
            )
            self.assertEqual(results.exit_code, 0)
            config = MappingConfig.load_mapping_config_from_path(confpath)
            self.assertEqual(len(config.mappings), 2)

    def test_generate_mapping_errors(self):
        """Exercise various error situations with the g2p generate-mapping CLI command"""

        results = self.runner.invoke(generate_mapping)
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("Nothing to do", results.output)

        results = self.runner.invoke(generate_mapping, "--ipa")
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("Missing argument", results.output)

        results = self.runner.invoke(generate_mapping, "fra")
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn(
            "Nothing to do",
            results.output,
            '"g2p generate-mapping fra" should say need --ipa or --dummy or --list-dummy',
        )

        results = self.runner.invoke(generate_mapping, "--ipa foo")
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("Invalid value for IN_LANG", results.output)

        results = self.runner.invoke(generate_mapping, "--dummy fra foo")
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("Invalid value for OUT_LANG", results.output)

        results = self.runner.invoke(generate_mapping, "--ipa crl")
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("Cannot find IPA mapping", results.output)

        results = self.runner.invoke(generate_mapping, "--ipa fra dan-ipa")
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("Cannot find IPA mapping", results.output)

        results = self.runner.invoke(generate_mapping, "--list-dummy")
        self.assertEqual(results.exit_code, 0)  # this one not an error
        self.assertIn("Dummy phone inventory", results.output)

        results = self.runner.invoke(generate_mapping, "--list-dummy fra")
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("IN_LANG is not allowed with --list-dummy", results.output)

        results = self.runner.invoke(generate_mapping, "--ipa --dummy fra")
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("Error: Multiple modes selected", results.output)

        results = self.runner.invoke(
            generate_mapping, "--out-dir does-not-exist --ipa fra"
        )
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn(
            "does not exist",
            results.output,
            "Non-existent out-dir must be reported as error",
        )

        results = self.runner.invoke(generate_mapping, "--from asdf")
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("Error: --from and --to must be used together", results.output)

        results = self.runner.invoke(
            generate_mapping, "--from fra_to_fra-ipa --to haa_to_haa-equiv"
        )
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("Cannot guess in/out for IPA lang spec", results.output)

        results = self.runner.invoke(generate_mapping, "--from eng --to fra[out]")
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("is only supported with the full", results.output)

        results = self.runner.invoke(
            generate_mapping, "--from fra_to_fra-ipa[foo] --to eng"
        )
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("is allowed in square brackets", results.output)

        results = self.runner.invoke(generate_mapping, "--from fra_to_eng --to eng")
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("Cannot find mapping", results.output)

        results = self.runner.invoke(generate_mapping, "--merge --from fra --to eng")
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn(
            "--merge is only compatible with --ipa and --dummy", results.output
        )

        results = self.runner.invoke(generate_mapping, "--merge --ipa fra")
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("OUT_LANG is required with --merge", results.output)

        results = self.runner.invoke(
            generate_mapping, "--ipa --out-dir foo_bar_baz fra"
        )
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("Invalid value for '--out-dir': Directory", results.output)

    def test_show_mappings(self):
        # One arg = all mappings to or from that language
        results = self.runner.invoke(show_mappings, ["fra-ipa", "--verbose"])
        self.assertEqual(results.exit_code, 0)
        self.assertIn("French to IPA", results.output)
        self.assertIn("French IPA to English IPA", results.output)
        self.assertEqual(len(re.findall(r"display_name", results.output)), 2)

        # One arg = all mappings to or from that language, terse output
        results = self.runner.invoke(show_mappings, ["fra-ipa"])
        self.assertEqual(results.exit_code, 0)
        self.assertIn("fra-ipa", results.output)
        self.assertIn("eng-ipa", results.output)
        self.assertEqual(len(re.findall(r"→", results.output)), 2)

        # Two conencted args = that mapping
        results = self.runner.invoke(show_mappings, ["fra", "fra-ipa", "--verbose"])
        self.assertEqual(results.exit_code, 0)
        self.assertIn("French to IPA", results.output)
        self.assertIn(r'{"in": "&", "out": "et"},', results.output)
        self.assertIn(
            r'{"in": "c", "out": "s", "context_after": "e|i|è|é|ê|ë|î|ï|ÿ"},',
            results.output,
        )
        self.assertIn(
            r'{"in": "e", "out": "", "context_before": "\\S", "context_after": "\\b"},',
            results.output,
        )
        self.assertEqual(len(re.findall(r"display_name", results.output)), 1)

        # Two args connected via a intermediate steps = all mappings on that path
        results = self.runner.invoke(show_mappings, ["fra", "eng-arpabet", "--verbose"])
        self.assertEqual(results.exit_code, 0)
        self.assertIn("French to IPA", results.output)
        self.assertIn("French IPA to English IPA", results.output)
        self.assertIn("English IPA to Arpabet", results.output)
        self.assertEqual(len(re.findall(r"display_name", results.output)), 3)

        # --all = all mappings
        results = self.runner.invoke(show_mappings, [])
        self.assertEqual(results.exit_code, 0)
        self.assertGreater(len(re.findall(r"→", results.output)), 100)

        # --csv = CSV formatted output
        results = self.runner.invoke(show_mappings, ["--csv", "crl-equiv", "--verbose"])
        self.assertEqual(results.exit_code, 0)
        self.assertIn("Northern East Cree Equivalencies", results.output)
        self.assertIn("thwaa,ᕨ,,", results.output)
        self.assertIn("Northern East Cree to IPA", results.output)
        self.assertIn("ᐧᕓ,vʷeː,,", results.output)

        # Bad language code
        results = self.runner.invoke(show_mappings, ["not-a-lang"])
        self.assertNotEqual(results.exit_code, 0)
        results = self.runner.invoke(show_mappings, ["fra", "not-a-lang"])
        self.assertNotEqual(results.exit_code, 0)

        # No path
        results = self.runner.invoke(show_mappings, ["fra", "moe"])
        self.assertNotEqual(results.exit_code, 0)

    def test_convert_from_file(self):
        input_file = os.path.join(DATA_DIR, "fra_simple.txt")
        results = self.runner.invoke(convert, [input_file, "fra", "fra-ipa", "--file"])
        self.assertEqual(results.exit_code, 0)
        self.assertIn("fʁɑ̃sɛ", results.output)
        with open(input_file, "r", encoding="utf8") as f:
            lines_in = len(list(f))
        # Make sure there is no resource warning about unclosed files
        self.assertNotIn("ResourceWarning", results.output)
        self.assertNotIn("unclosed file", results.output)
        # The output should have the same number of lines as the input
        self.assertEqual(lines_in, len(results.output.splitlines()))

        # - is stdin
        results = self.runner.invoke(
            convert, ["--file", "-", "fra", "fra-ipa"], input="français"
        )
        self.assertEqual(results.exit_code, 0)
        self.assertIn("fʁɑ̃sɛ", results.output)

        # warning about deprecated heuristic file detection
        with self.assertLogs(LOGGER, "WARNING") as cm:
            self.runner.invoke(convert, [input_file, "fra", "fra-ipa"])
        self.assertIn("deprecated", "".join(cm.output))

        # Error for --file with non existent file
        results = self.runner.invoke(
            convert, ["does_not_exist.txt", "fra", "fra-ipa", "--file"]
        )
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("No such file or directory", results.output)

    def test_convert_errors(self):
        """Exercise code handling error situations in g2p convert"""
        results = self.runner.invoke(convert, "asdf bad_in_lang eng-ipa")
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("not a valid value for 'IN_LANG'", results.output)

        results = self.runner.invoke(convert, "asdf fra bad_out_lang")
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("not a valid value for 'OUT_LANG'", results.output)

        results = self.runner.invoke(convert, "asdf fra dan")
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("Path between", results.output)
        self.assertIn("does not exist", results.output)

        results = self.runner.invoke(
            convert, "--no-tok --tok-lang fra asdf fra fra-ipa"
        )
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn(
            "Specified conflicting --no-tok and --tok-lang options", results.output
        )

    def test_short_dash_h(self):
        results_short = self.runner.invoke(convert, "-h")
        self.assertEqual(results_short.exit_code, 0)
        self.assertIn("Show this message and exit", results_short.output)
        results_long = self.runner.invoke(convert, "--help")
        self.assertEqual(results_long.exit_code, 0)
        self.assertEqual(results_short.output, results_long.output)

    def test_generate_mapping(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            result = self.runner.invoke(
                generate_mapping, ["--ipa", "--out-dir", tmpdir, "fra"]
            )
            self.assertEqual(result.exit_code, 0)
            with open(
                os.path.join(tmpdir, "fra-ipa_to_eng-ipa.json"), "r", encoding="utf8"
            ) as f:
                fra2eng_ipa = json.load(f)
            for s in ("ɛj", "ks", "ɔn"):
                self.assertIn({"in": s, "out": s}, fra2eng_ipa)


if __name__ == "__main__":
    main()



from collections import deque
from typing import (
    Any,
    Deque,
    Dict,
    Generic,
    Hashable,
    Iterable,
    Iterator,
    List,
    Set,
    Tuple,
    TypeVar,
    Union,
)

from typing_extensions import TypedDict

T = TypeVar("T", bound=Hashable)


class DiGraph(Generic[T]):
    """A simple directed graph class

    Most functions raise KeyError if called with a node u or v not in the graph.
    """

    def __init__(self) -> None:
        """Contructor, empty if no data, else load from data"""
        self._edges: Dict[T, List[T]] = {}

    def clear(self) -> None:
        """Clear the graph"""
        self._edges.clear()

    def update(self, edges: Iterable[Tuple[T, T]], nodes: Iterable[T]):
        """Update the graph with new edges and nodes"""
        for node in nodes:
            self.add_node(node)
        for u, v in edges:
            self.add_edge(u, v)

    def add_node(self, u: T):
        """Add a node to the graph"""
        if u not in self._edges:
            self._edges[u] = []

    def add_edge(self, u: T, v: T):
        """Add a directed edge from u to v"""
        self.add_node(u)
        self.add_node(v)
        if v not in self._edges[u]:
            self._edges[u].append(v)

    def add_edges_from(self, edges: Iterable[Tuple[T, T]]):
        """Add edges from a list of tuples"""
        for u, v in edges:
            self.add_edge(u, v)

    @property  # read-only
    def nodes(self):
        """Return the nodes"""
        return self._edges.keys()

    @property  # read-only
    def edges(self) -> Iterator[Tuple[T, T]]:
        """Iterate over all edges"""
        for u, neighbours in self._edges.items():
            for v in neighbours:
                yield u, v

    def __contains__(self, u: T) -> bool:
        """Check if a node is in the graph"""
        return u in self._edges

    def has_path(self, u: T, v: T) -> bool:
        """Check if there is a path from u to v"""
        if v not in self._edges:
            raise KeyError(f"Node {v} not in graph")
        visited: Set[T] = set()
        return self._has_path(u, v, visited)

    def _has_path(self, u: T, v: T, visited: Set[T]) -> bool:
        """Helper function for has_path"""
        visited.add(u)
        if u == v:
            return True
        for neighbour in self._edges[u]:
            if neighbour not in visited:
                if self._has_path(neighbour, v, visited):
                    return True
        return False

    def successors(self, u: T) -> Iterator[T]:
        """Return the successors of u"""
        return iter(self._edges[u])

    def descendants(self, u: T) -> Set[T]:
        """Return the descendants of u"""
        visited: Set[T] = set()
        self._descendants(u, visited)
        visited.remove(u)
        return visited

    def _descendants(self, u: T, visited: Set[T]):
        """Helper function for descendants"""
        visited.add(u)
        for neighbour in self._edges[u]:
            if neighbour not in visited:
                self._descendants(neighbour, visited)

    def ancestors(self, u: T) -> Set[T]:
        """Return the ancestors of u"""
        reversed_graph: DiGraph[T] = DiGraph()
        reversed_graph.add_edges_from((v, u) for u, v in self.edges)
        for node in self.nodes:
            reversed_graph.add_node(node)
        return reversed_graph.descendants(u)

    def shortest_path(self, u: T, v: T) -> List[T]:
        """Return the shortest path from u to v

        Algorithm: Dijsktra's algorithm for unweighted graphs, which is just BFS

        Returns:
            list: the shortest path from u to v
        Raises:
            KeyError: if u or v is not in the graph
            ValueError: if there is no path from u to v
        """

        if v not in self._edges:
            raise KeyError(f"Node {v} not in graph")
        visited: Dict[T, Union[T, None]] = {
            u: None
        }  # dict of {node: predecessor on shortest path from u}
        queue: Deque[T] = deque([u])
        while queue:
            u = queue.popleft()
            if u == v:
                rev_path: List[T] = []
                nextu: Union[T, None] = u
                while nextu is not None:
                    rev_path.append(nextu)
                    nextu = visited[nextu]
                return list(reversed(rev_path))
            for neighbour in self._edges[u]:
                if neighbour not in visited:
                    visited[neighbour] = u
                    queue.append(neighbour)
        raise ValueError(f"No path from {u} to {v}")


NodeDict = TypedDict("NodeDict", {"id": Any})


class NodeLinkDict(TypedDict, Generic[T]):
    source: T
    target: T


class NodeLinkDataDict(TypedDict, Generic[T]):
    directed: bool
    graph: Dict
    links: List[NodeLinkDict[T]]
    multigraph: bool
    nodes: List[NodeDict]


def node_link_graph(data: NodeLinkDataDict[T]) -> DiGraph[T]:
    """Replacement for networkx.node_link_graph"""
    if not data.get("directed", False):
        raise ValueError("Graph must be directed")
    if data.get("multigraph", True):
        raise ValueError("Graph must not be a multigraph")
    if not isinstance(data.get("nodes", None), list):
        raise ValueError('data["nodes"] must be a list')
    if not isinstance(data.get("links", None), list):
        raise ValueError('data["links"] must be a list')

    graph: DiGraph[T] = DiGraph()
    for node in data["nodes"]:
        graph.add_node(node["id"])
    for edge in data["links"]:
        graph.add_edge(edge["source"], edge["target"])
    return graph


def node_link_data(graph: DiGraph[T]) -> NodeLinkDataDict[T]:
    """Replacement for networkx.node_link_data"""
    nodes: List[NodeDict] = [{"id": node} for node in graph.nodes]
    links: List[NodeLinkDict[T]] = [{"source": u, "target": v} for u, v in graph.edges]
    return {
        "directed": True,
        "graph": {},
        "links": links,
        "multigraph": False,
        "nodes": nodes,
    }



#!/usr/bin/env python

import gzip
import json
from typing import Any
from unittest import TestCase, main

from g2p import make_g2p
from g2p.exceptions import InvalidLanguageCode, NoPath
from g2p.log import LOGGER
from g2p.mappings.langs import LANGS_NWORK_PATH
from g2p.mappings.langs.network_lite import DiGraph, node_link_data, node_link_graph
from g2p.transducer import CompositeTransducer, Transducer


class NetworkTest(TestCase):
    """Basic Test for available networks"""

    def setUp(self):
        pass

    def test_not_found(self):
        with self.assertRaises(InvalidLanguageCode):
            with self.assertLogs(LOGGER, level="ERROR"):
                make_g2p("foo", "eng-ipa")
        with self.assertRaises(InvalidLanguageCode):
            with self.assertLogs(LOGGER, level="ERROR"):
                make_g2p("git", "bar")

    def test_no_path(self):
        with self.assertRaises(NoPath), self.assertLogs(LOGGER, level="ERROR"):
            make_g2p("hei", "git")

    def test_valid_composite(self):
        transducer = make_g2p("atj", "eng-ipa", tokenize=False)
        self.assertTrue(isinstance(transducer, CompositeTransducer))
        self.assertEqual("niɡiɡw", transducer("nikikw").output_string)

    def test_valid_transducer(self):
        transducer = make_g2p("atj", "atj-ipa", tokenize=False)
        self.assertTrue(isinstance(transducer, Transducer))
        self.assertEqual("niɡiɡw", transducer("nikikw").output_string)


class NetworkLiteTest(TestCase):
    data: Any

    @classmethod
    def setUpClass(cls):
        with gzip.open(LANGS_NWORK_PATH, "rt", encoding="utf8") as f:
            cls.data = json.load(f)

    def test_has_path(self):
        graph: DiGraph = DiGraph()
        graph.add_edge("a", "b")
        graph.add_edge("b", "a")  # cycle
        graph.add_edge("a", "c")
        graph.add_edge("c", "d")
        graph.add_edge("e", "f")
        self.assertTrue(graph.has_path("a", "c"))
        self.assertTrue(graph.has_path("a", "d"))
        self.assertTrue(graph.has_path("b", "a"))
        self.assertFalse(graph.has_path("a", "e"))
        self.assertFalse(graph.has_path("a", "f"))
        self.assertFalse(graph.has_path("c", "a"))
        with self.assertRaises(KeyError):
            graph.has_path("a", "y")
        with self.assertRaises(KeyError):
            graph.has_path("x", "b")

    def test_g2p_path(self):
        graph = node_link_graph(self.data)
        self.assertTrue(graph.has_path("atj", "eng-ipa"))
        self.assertTrue(graph.has_path("atj", "atj-ipa"))
        self.assertFalse(graph.has_path("hei", "git"))

    def test_successors(self):
        graph: DiGraph = DiGraph()
        graph.add_edge("a", "b")
        graph.add_edge("b", "a")
        graph.add_edge("a", "c")
        self.assertEqual(set(graph.successors("a")), {"b", "c"})
        self.assertEqual(set(graph.successors("b")), {"a"})
        self.assertEqual(set(graph.successors("c")), set())

    def test_descendants(self):
        graph: DiGraph = DiGraph()
        graph.add_edge("a", "b")
        graph.add_edge("b", "a")  # cycle
        graph.add_edge("a", "c")
        graph.add_edge("c", "d")
        graph.add_edge("e", "f")
        self.assertEqual(graph.descendants("a"), {"b", "c", "d"})
        self.assertEqual(graph.descendants("b"), {"a", "c", "d"})
        self.assertEqual(graph.descendants("c"), {"d"})
        self.assertEqual(graph.descendants("d"), set())
        self.assertEqual(graph.descendants("e"), {"f"})
        self.assertEqual(graph.descendants("f"), set())
        with self.assertRaises(KeyError):
            graph.descendants("x")

    def test_g2p_descendants(self):
        graph = node_link_graph(self.data)
        self.assertEqual(
            graph.descendants("atj"), {"eng-ipa", "atj-ipa", "eng-arpabet"}
        )
        self.assertEqual(graph.descendants("eng-ipa"), {"eng-arpabet"})
        self.assertEqual(graph.descendants("atj-ipa"), {"eng-ipa", "eng-arpabet"})
        self.assertEqual(graph.descendants("eng-arpabet"), set())

    def test_ancestors(self):
        graph: DiGraph = DiGraph()
        graph.add_edge("a", "b")
        graph.add_edge("a", "c")
        graph.add_edge("d", "a")  # cycle
        graph.add_edge("c", "d")
        graph.add_edge("e", "f")
        self.assertEqual(graph.ancestors("a"), {"c", "d"})
        self.assertEqual(graph.ancestors("b"), {"a", "d", "c"})
        self.assertEqual(graph.ancestors("c"), {"a", "d"})
        self.assertEqual(graph.ancestors("d"), {"a", "c"})
        self.assertEqual(graph.ancestors("e"), set())
        self.assertEqual(graph.ancestors("f"), {"e"})
        with self.assertRaises(KeyError):
            graph.ancestors("x")

    def test_g2p_ancestors(self):
        graph: DiGraph = node_link_graph(self.data)
        self.assertEqual(graph.ancestors("atj"), set())
        self.assertGreater(len(graph.ancestors("eng-ipa")), 50)

    def test_shortest_path(self):
        graph: DiGraph = DiGraph()
        graph.add_edge("a", "e")
        graph.add_edge("e", "f")
        graph.add_edge("f", "g")
        graph.add_edge("g", "d")
        graph.add_edge("f", "d")
        graph.add_edge("a", "b")
        graph.add_edge("b", "a")  # Cycle
        graph.add_edge("a", "c")
        graph.add_edge("c", "d")
        graph.add_edge("a", "d")
        graph.add_edge("b", "d")
        self.assertEqual(graph.shortest_path("a", "d"), ["a", "d"])
        self.assertEqual(graph.shortest_path("c", "d"), ["c", "d"])
        self.assertEqual(graph.shortest_path("a", "a"), ["a"])
        with self.assertRaises(ValueError):
            graph.shortest_path("c", "a")
        with self.assertRaises(KeyError):
            graph.shortest_path("a", "y")
        with self.assertRaises(KeyError):
            graph.shortest_path("x", "b")

    def test_g2p_shortest_path(self):
        graph = node_link_graph(self.data)
        self.assertEqual(
            graph.shortest_path("atj", "eng-arpabet"),
            ["atj", "atj-ipa", "eng-ipa", "eng-arpabet"],
        )

    def test_contains(self):
        graph: DiGraph = DiGraph()
        graph.add_edge("a", "b")
        self.assertTrue("a" in graph)
        self.assertTrue("b" in graph)
        self.assertFalse("c" in graph)

    def test_node_link_data(self):
        graph = node_link_graph(self.data)
        self.assertEqual(node_link_data(graph), self.data)

    def test_node_link_graph_errors(self):
        with self.assertRaises(ValueError):
            node_link_graph({**self.data, "directed": False})  # type: ignore
        with self.assertRaises(ValueError):
            node_link_graph({**self.data, "multigraph": True})  # type: ignore
        with self.assertRaises(ValueError):
            node_link_graph({**self.data, "nodes": "not a list"})  # type: ignore
        with self.assertRaises(ValueError):
            node_link_graph({**self.data, "links": "not a list"})  # type: ignore
        with self.assertRaises(ValueError):
            data = self.data.copy()
            del data["nodes"]
            node_link_graph(data)
        with self.assertRaises(ValueError):
            data = self.data.copy()
            del data["links"]
            node_link_graph(data)

    def test_no_duplicates(self):
        graph: DiGraph = DiGraph()
        graph.add_edge("a", "b")
        graph.add_edge("b", "c")
        graph.add_edge("a", "c")
        graph.add_edge("a", "b")
        self.assertEqual(len(list(graph.edges)), 3)
        self.assertEqual(len(graph.nodes), 3)
        self.assertEqual(len(list(graph.successors("a"))), 2)
        self.assertEqual(len(list(graph.successors("b"))), 1)
        self.assertEqual(len(list(graph.successors("c"))), 0)


if __name__ == "__main__":
    main()



"""
Command line interface to the g2p system
"""

import io
import json
import os
import pprint
import re
import sys
from pathlib import Path
from textwrap import dedent
from typing import List, Tuple

import click

import g2p._version
from g2p import make_g2p, make_tokenizer
from g2p.constants import (
    DISTANCE_METRICS,
    LANGS_DIR,
    LANGS_FILE_NAME,
    NETWORK_FILE_NAME,
)
from g2p.exceptions import InvalidLanguageCode, MappingMissing, NoPath
from g2p.static import __file__ as static_file

PRINTER = pprint.PrettyPrinter(indent=4)

if "pytest" not in sys.modules:  # pragma: no cover
    if sys.stdout.encoding != "utf8" and hasattr(sys.stdout, "buffer"):
        sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding="utf8")
    if sys.stderr.encoding != "utf8" and hasattr(sys.stderr, "buffer"):
        sys.stderr = io.TextIOWrapper(sys.stderr.buffer, encoding="utf8")


CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


def parse_from_or_to_lang_spec(lang_spec):
    """Parse a value given to g2p generate-mapping --from or --to.

    See the documentation of generate_mapping() for the syntax of lang_spec.

    Returns list[tuple[Mapping, io (str)]]:
        the mapping(s) lang_spec refers to, and "in" or "out", to indicate if the
        relevant inventory is the mapping's in_lang or out_lang.

    Raises:
        click.BadParameter if lang_spec is not valid
    """
    # Defer expensive imports
    from g2p.mappings import MAPPINGS_AVAILABLE, Mapping
    from g2p.mappings.utils import is_ipa

    mapping_spec, _, in_or_out = lang_spec.partition("[")
    in_or_out.rstrip("]")
    in_lang, _, out_lang = mapping_spec.partition("_to_")

    if out_lang:
        try:
            mapping = Mapping.find_mapping(in_lang=in_lang, out_lang=out_lang)
        except MappingMissing as e:
            raise click.BadParameter(
                f'Cannot find mapping {in_lang}->{out_lang} for --from or --to spec "{lang_spec}": {e}'
            )
        if not in_or_out:
            if is_ipa(out_lang):
                in_or_out = "out"
            elif is_ipa(in_lang):
                in_or_out = "in"
            else:
                raise click.BadParameter(
                    f'Cannot guess in/out for IPA lang spec "{lang_spec}" because neither {in_lang} '
                    f'nor {out_lang} is IPA. Specify "[in]" or "[out]" if you are sure it is correct.'
                )
        if in_or_out not in ("in", "out"):
            raise click.BadParameter(
                f'Invalid IPA language specification "{lang_spec}": only "in" or "out" '
                "is allowed in square brackets, to disambiguate between input or output "
                "inventory when necessary."
            )
        return [(mapping, in_or_out)]

    else:
        if in_or_out:
            raise click.BadParameter(
                f'Bad IPA lang spec "{lang_spec}": the [in]/[out] qualifier is only '
                "supported with the full in-lang_to_out-lang[[in]|[out]] syntax."
            )
        if in_lang == "eng":
            mapping = Mapping.find_mapping(in_lang="eng-ipa", out_lang="eng-arpabet")
            in_or_out = "in"
            return [(mapping, in_or_out)]
        else:
            out_lang = in_lang + "-ipa"
            # check_ipa_known_segs([out_lang])  # this outputs a lot of spurious noise...
            mappings = [
                (Mapping.find_mapping(in_lang=m.in_lang, out_lang=m.out_lang), "out")
                for m in MAPPINGS_AVAILABLE
                if m.out_lang == out_lang and not is_ipa(m.in_lang)
            ]
            if not mappings:
                raise click.BadParameter(f'No IPA mappings found for "{lang_spec}".')
            return mappings


@click.version_option(version=g2p._version.VERSION, prog_name="g2p")
@click.group(context_settings=CONTEXT_SETTINGS)
def cli():
    """Management script for G2P"""


@click.option(
    "--out-dir",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    help='Output results in DIRECTORY instead of the global "generated" directory.',
)
@click.option(
    "--to",
    "to_langs",
    default=None,
    help='Colon- or comma-separated list of "to" languages in from/to mode',
)
@click.option(
    "--from",
    "from_langs",
    default=None,
    help='Colon- or comma-separated list of "from" languages in from/to mode',
)
@click.option(
    "--list-dummy", default=False, is_flag=True, help="List the dummy phone inventory."
)
@click.option(
    "--dummy/--no-dummy",
    default=False,
    help="Generate dummy fallback mapping to minimalist phone inventory.",
)
@click.option(
    "--ipa/--no-ipa", default=False, help="Generate mapping from LANG-ipa to eng-ipa."
)
@click.option(
    "--merge/--no-merge",
    default=False,
    help="Merge multiple mappings together, in which case IN_LANG is a colon-seperated list and OUT_LANG is required.",
)
@click.argument("out_lang", required=False, default=None, type=str)
@click.argument("in_lang", required=False, default=None, type=str)
@cli.command(
    context_settings=CONTEXT_SETTINGS,
    short_help="Generate English IPA or dummy mapping.",
)
@click.option(
    "--distance",
    type=click.Choice(DISTANCE_METRICS),
    required=False,
    default="weighted_feature_edit_distance",
)
def generate_mapping(  # noqa: C901
    in_lang,
    out_lang,
    dummy,
    ipa,
    list_dummy,
    out_dir,
    merge,
    from_langs,
    to_langs,
    distance,
) -> None:
    """Generate a new mapping from existing mappings in the g2p system.

    This command has different modes of operation.

    Standard mode:

      g2p generate-mapping [--dummy|--ipa] IN_LANG [OUT_LANG]

      For specified IN_LANG, generate a mapping from IN_LANG-ipa to eng-ipa,
    or from IN_LANG-ipa to a dummy minimalist phone inventory. This assumes
    the mapping IN_LANG -> IN_LANG-ipa exists and creates a mapping from its
    output inventory.

      To generate a mapping from IN_LANG-ipa to eng-ipa from a mapping
    following a different patterns, e.g., from crl-equiv -> crl-ipa, specify
    both IN_LANG (crl-equiv in this example) and OUT_LANG (crl-ipa in this
    example).

      \b
      Sample usage:
        Generate Algonquin IPA to English IPA from alq -> alq-ipa:
            g2p generate-mapping --ipa alq
        Generate Mohawk IPA to English IPA from moh-equiv -> moh-ipa:
            g2p generate-mapping --ipa moh-equiv moh-ipa
        Generate Michif IPA to English IPA from the union of crg-dv -> crg-ipa
        and crg-tmd -> crg-ipa:
            g2p generate-mapping --ipa --merge crg-dv:crg-tmd crg-ipa

    List the dummy inventory used by --dummy:

      g2p generate-mapping --list-dummy

    From/to IPA mode:

    \b
      g2p generate-mapping --from FROM_L1 --to TO_L1
      g2p generate-mapping --from FROM_L1:FROM_L2:... --to TO_L1:TO_L2:...

      Generate an IPA mapping from the union of FROM_L1-ipa, FROM-L2-ipa, etc to
    the union of TO_L1-ipa, TO-L2-ipa, etc. One or more from/to language
    code(s) can be specified in colon- or comma-separated lists. Note, by default
    we use Panphon's weighted_feature_edit_distance, but you can change this with
    the --distance argument

    \b
      Sample usage:
        Generate a mapping from kwk-ipa to moh-ipa based on all mappings into
        kwk-ipa and moh-ipa:
            g2p generate-mapping --from kwk --to moh
        Generate a mapping from eng-ipa to crg-ipa based only on crg-dv -> crg-ipa:
            g2p generate-mapping --from eng --to crg-dv_to_crg-ipa
        Generate a mapping from kwk-ipa to moh-ipa+crg-ipa+eng-ipa based on
        all mappings into kwk-ipa (from side) and the union of all mappings
        into moh-ipa and crg-ipa plus eng-ipa_to_eng-arpabet (to side):
            g2p generate-mapping --from kwk --to moh:crg:eng

      Full syntax for specifying FROM_Ln and TO_Ln:

      \b
        lang (i.e., 3-letter code):
         - If there is only one mapping into lang-ipa, "lang" refers to the
           output of that mapping, e.g., "fra" means "fra_to_fra-ipa[out]".
         - If there are several mappings into lang-ipa, "lang" refers to the
           union of the outputs of those mappings, e.g., "moh" means the union
           of "moh-equiv_to_moh-ipa[out]" and "moh-festival_to_moh-ipa[out]".
         - It is an error if there are no mappings into lang-ipa.
         - Only mappings from non-IPA to IPA are considered (i.e., IPA-to-IPA
           mappings created by this command will not be included: use the
           longer syntax below if you want to use them).
         - Special case: "eng" refers to "eng-ipa_to_eng-arpabet[in]".

      \b
        in-lang_to_out-lang[[in]|[out]]:
         - This expanded syntax is used to avoid the union when it is not
           desired, e.g., "moh-equiv_to_moh-ipa" refers only to
           "moh-equiv_to_moh-ipa,out" rather than the union "moh" represents.
         - If out-lang is IPA, the output inventory is used; else if in-lang
           is IPA, the input inventory is used; it is an error if neither
           language is IPA.
         - Specify "[in]" or "[out]" to override the above default.
         - "_to_" is the joiner used to specify "the mapping from 'in-lang' to
           'out-lang'" in the g2p network, regardless of the name of the file
           it is stored in.

    If you just modified or created the mappings from which the new mapping is
    to be generated, don't forget to call "g2p update" first, so that "g2p
    generate-mapping" can see the latest version.

    Call "g2p update" again after calling "g2p generate-mapping" to compile
    the newly generated mapping and make it available.

    Note: exactly one of --ipa, --dummy, --from/--to, or --list-dummy is
    required.

    You can list available mappings with "g2p doctor --list-ipa", or by
    visiting http://g2p-studio.herokuapp.com/api/v1/langs .
    """
    # Defer expensive imports
    from g2p.log import LOGGER
    from g2p.mappings import Mapping
    from g2p.mappings.create_fallback_mapping import (
        DUMMY_INVENTORY,
        align_to_dummy_fallback,
    )
    from g2p.mappings.create_ipa_mapping import create_mapping, create_multi_mapping
    from g2p.mappings.langs import LANGS_NETWORK
    from g2p.mappings.langs.utils import check_ipa_known_segs
    from g2p.mappings.utils import is_ipa, is_xsampa

    # Make sure only one mode was specified on the command line
    mode_count = (
        (1 if ipa else 0)
        + (1 if dummy else 0)
        + (1 if list_dummy else 0)
        + (1 if (from_langs or to_langs) else 0)
    )
    if mode_count == 0:
        raise click.UsageError(
            "Nothing to do! Please specify at least one of --ipa, --dummy, "
            "--list-dummy, or --from/--to."
        )
    if mode_count > 1:
        raise click.UsageError(
            "Multiple modes selected. Choose only one of --ipa, --dummy, "
            "--list-dummy, or --from/--to."
        )

    if list_dummy or from_langs is not None or to_langs is not None:
        if in_lang is not None:
            raise click.UsageError(
                "IN_LANG is not allowed with --list-dummy or --from/--too",
            )

    if from_langs is not None or to_langs is not None:
        if from_langs is None or to_langs is None:
            raise click.UsageError("--from and --to must be used together")

    if merge:
        if not ipa and not dummy:
            raise click.UsageError("--merge is only compatible with --ipa and --dummy.")
        if out_lang is None:
            raise click.UsageError("OUT_LANG is required with --merge.")

    if list_dummy:
        # --list-dummy mode
        print(f"Dummy phone inventory: {DUMMY_INVENTORY}")

    elif ipa or dummy:
        # --ipa and --dummy modes
        if in_lang is None:
            raise click.UsageError("Missing argument 'IN_LANG'.")
        if merge:
            in_langs = in_lang.split(":")
        else:
            in_langs = [in_lang]

        in_lang_choices = [
            x for x in LANGS_NETWORK.nodes if not is_ipa(x) and not is_xsampa(x)
        ]
        for in_lang in in_langs:
            if in_lang not in in_lang_choices:
                raise click.UsageError(
                    f'Invalid value for IN_LANG: "{in_lang}".\n'
                    "IN_LANG must be a non-IPA language code with an existing IPA mapping, "
                    f"i.e., one of:\n{', '.join(in_lang_choices)}."
                )

        out_lang_choices = [x for x in LANGS_NETWORK.nodes if is_ipa(x)]
        if out_lang is None:
            out_lang = f"{in_lang}-ipa"
        elif out_lang not in out_lang_choices:
            raise click.UsageError(
                f'Invalid value for OUT_LANG: "{out_lang}".\n'
                "OUT_LANG must be an IPA language code with an existing mapping from IN_LANG, "
                f"i.e., one of:\n{', '.join(out_lang_choices)}"
            )

        source_mappings = []
        for in_lang in in_langs:
            try:
                source_mapping = Mapping.find_mapping(
                    in_lang=in_lang, out_lang=out_lang
                )
            except MappingMissing as e:
                raise click.BadParameter(
                    f'Cannot find IPA mapping from "{in_lang}" to "{out_lang}": {e}',
                    param_hint=("IN_LANG", "OUT_LANG"),  # type: ignore
                )
            source_mappings.append(source_mapping)

        if ipa:
            check_ipa_known_segs([f"{in_lang}-ipa"])
            eng_ipa = Mapping.find_mapping(in_lang="eng-ipa", out_lang="eng-arpabet")
            click.echo(f"Writing English IPA mapping for {out_lang} to file")
            new_mapping = create_mapping(source_mappings[0], eng_ipa, distance=distance)
            for m in source_mappings[1:]:
                new_mapping.extend(create_mapping(m, eng_ipa, distance=distance))
        else:  # dummy
            click.echo(f"Writing dummy fallback mapping for {out_lang} to file")
            new_mapping = align_to_dummy_fallback(source_mappings[0], distance=distance)
            for m in source_mappings[1:]:
                new_mapping.extend(align_to_dummy_fallback(m, distance=distance))

        new_mapping.deduplicate()

        if out_dir:
            new_mapping.config_to_file(out_dir)
            new_mapping.mapping_to_file(out_dir)
        else:
            new_mapping.config_to_file()
            new_mapping.mapping_to_file()

    elif from_langs is not None:
        # --from/--to mode
        assert to_langs is not None

        from_mappings: List[Tuple[Mapping, str]] = []
        for from_lang in re.split(r"[:,]", from_langs):
            from_mappings.extend(parse_from_or_to_lang_spec(from_lang))
        to_mappings: List[Tuple[Mapping, str]] = []
        for to_lang in re.split(r"[:,]", to_langs):
            to_mappings.extend(parse_from_or_to_lang_spec(to_lang))

        if not from_mappings:
            raise click.UsageError(
                f'Invalid --from value "{from_langs}": no mappings found.'
            )
        if not to_mappings:
            raise click.UsageError(
                f'Invalid --to value "{to_langs}": no mappings found.'
            )

        for from_mapping, in_or_out in from_mappings:
            LOGGER.info(
                f"From mapping: {from_mapping.in_lang}_to_{from_mapping.out_lang}[{in_or_out}]"
            )
        for to_mapping, in_or_out in to_mappings:
            LOGGER.info(
                f"To mapping: {to_mapping.in_lang}_to_{to_mapping.out_lang}[{in_or_out}]"
            )

        new_mapping = create_multi_mapping(
            from_mappings, to_mappings, distance=distance
        )

        if out_dir:
            new_mapping.config_to_file(out_dir)
            new_mapping.mapping_to_file(out_dir)
        else:
            new_mapping.config_to_file()
            new_mapping.mapping_to_file()


@click.option(
    "--substring-alignments",
    "-a",
    default=False,
    is_flag=True,
    help="Show the minimal monotonic substring alignments.",
)
@click.option(
    "--pretty-edges",
    "-e",
    default=False,
    is_flag=True,
    help="Show the traduction graph in a pretty, plain-text format.",
)
@click.option(
    "--debugger/--no-debugger",
    "-d",
    default=False,
    is_flag=True,
    help="Show all the conversion steps applied.",
)
@click.option(
    "--file",
    "-f",
    default=False,
    is_flag=True,
    help="Read input from plain-text file INPUT_TEXT.",
)
@click.option(
    "--check/--no-check",
    "-c",
    default=False,
    is_flag=True,
    help="Check IPA outputs against panphon and/or eng-arpabet output against ARPABET",
)
@click.option(
    "--neural/--no-neural",
    "-n",
    default=False,
    is_flag=True,
    help="Allow neural mappings if available. Requires installation of optional neural dependencies (pip install g2p[neural])",
)
@click.option(
    "--tok-lang",
    default=None,
    help="Override the tokenizing language. Implies --tok.",
)
@click.option(
    "--tok/--no-tok",
    "-t",
    default=None,  # three-way var: None=not set, True/False=set to True/False
    is_flag=True,
    help="Tokenize INPUT_TEXT before converting. Default is --tok, specify --no-tok to turn off.",
)
@click.option(
    "--config",
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    help="A path to a mapping configuration file to use",
)
@click.argument("out_lang")
@click.argument("in_lang")
@click.argument("input_text", type=click.STRING)
@cli.command(
    context_settings=CONTEXT_SETTINGS,
    short_help="Convert text through a g2p mapping path.",
)
def convert(  # noqa: C901
    in_lang,
    out_lang,
    input_text,
    tok,
    check,
    file,
    debugger,
    pretty_edges,
    neural,
    tok_lang,
    config,
    substring_alignments,
):
    """Convert INPUT_TEXT through g2p mapping(s) from IN_LANG to OUT_LANG.

    Visit http://g2p-studio.herokuapp.com/api/v1/langs for a list of languages.

    There must be a path from IN_LANG to OUT_LANG, possibly via some intermediates.
    For example, mapping from fra to eng-arpabet will successively apply
    fra->fra-ipa, fra-ipa->eng-ipa and eng-ipa->eng-arpabet.
    """
    # Defer expensive imports
    from g2p.log import LOGGER
    from g2p.mappings import MAPPINGS_AVAILABLE, Mapping, MappingConfig
    from g2p.mappings.langs import LANGS_NETWORK

    # Check valid input
    # Check input != output
    if in_lang == out_lang:
        raise click.UsageError("Values must be different for 'IN_LANG' and 'OUT_LANG'")
    if config:
        # This isn't that DRY - copied from g2p/mappings/langs/__init__.py
        mappings_legal_pairs = []
        mapping_config = MappingConfig.load_mapping_config_from_path(config)
        for index, mapping in enumerate(mapping_config.mappings):
            mappings_legal_pairs.append((mapping.in_lang, mapping.out_lang))
            mapping_config.mappings[index] = Mapping.load_mapping_from_path(
                config, index
            )
        for pair in mappings_legal_pairs:
            if pair[0] in LANGS_NETWORK.nodes:
                LOGGER.warning(
                    f"A mapping with the name '{pair[0]}' is already defined in g2p. "
                    "Your local mapping with the same name might not function properly."
                )
        LANGS_NETWORK.add_edges_from(mappings_legal_pairs)
        MAPPINGS_AVAILABLE.extend(mapping_config.mappings)
    # Check input lang exists
    if in_lang not in LANGS_NETWORK.nodes:
        raise click.UsageError(f"'{in_lang}' is not a valid value for 'IN_LANG'")
    # Check output lang exists
    if out_lang not in LANGS_NETWORK.nodes:
        raise click.UsageError(f"'{out_lang}' is not a valid value for 'OUT_LANG'")
    # Check if path exists
    if not LANGS_NETWORK.has_path(in_lang, out_lang):
        raise click.UsageError(
            f"Path between '{in_lang}' and '{out_lang}' does not exist"
        )
    to_close = None
    try:
        if file:
            if input_text == "-":
                lines = sys.stdin
            else:
                try:
                    to_close = lines = open(input_text, encoding="utf8")
                except FileNotFoundError as e:
                    raise click.UsageError(f"Could not open file {input_text}: {e}")
        else:
            lines = [input_text]
            if os.path.exists(input_text) and input_text.endswith(".txt"):
                LOGGER.warning(
                    "The old heuristic for detecting file input is deprecated, "
                    f"specify --file if you meant to read your input text from file {input_text}."
                )
        # Determine which tokenizer to use, if any
        if tok is not None and not tok and tok_lang is not None:
            raise click.UsageError(
                "Specified conflicting --no-tok and --tok-lang options."
            )
        if tok is None:
            tok = True  # Tokenize by default
        custom_tokenizer = make_tokenizer(tok_lang) if tok_lang else None
        # Transduce!!!
        assert in_lang and out_lang
        transducer = make_g2p(
            in_lang,
            out_lang,
            tokenize=tok,
            neural=neural,
            custom_tokenizer=custom_tokenizer,
        )
        no_tok_warning_printed = False
        for line in lines:
            if not tok and " " in line and not no_tok_warning_printed:
                LOGGER.warning(
                    "You disabled tokenization but your text contains spaces. --no-tok is intended for use when you are supplying isolated words; rules meant to apply at word boundaries might not work correctly."
                )
                no_tok_warning_printed = True
            tg = transducer(line)
            if check:
                transducer.check(tg, display_warnings=True)
            outputs = [tg.output_string]
            if substring_alignments:
                outputs += [tg.substring_alignments()]
            if pretty_edges:
                outputs += [tg.pretty_edges()]
            if debugger:
                outputs += [tg.edges, tg.debugger]
            if len(outputs) > 1:
                click.echo(pprint.pformat(outputs, indent=4))
            else:
                click.echo(tg.output_string, nl=not file)
    finally:
        if to_close is not None:
            to_close.close()


# Note: with -m eng-ipa, we actually check all the mappings from lang-ipa to eng-ipa.
@click.option(
    "--list-all", is_flag=True, help="List all mappings that can be specified"
)
@click.option(
    "--list-ipa", is_flag=True, help="List IPA mappings that can be specified"
)
@click.option(
    "--mapping",
    "-m",
    multiple=True,
    help="Check specified IPA mapping(s) (default: check all IPA mappings).",
)
@cli.command(context_settings=CONTEXT_SETTINGS)
def doctor(mapping, list_all, list_ipa):
    """Check for common errors in mappings.

    There should eventually be more checks here, but doctor currently checks for:

    1. Characters that are in IPA mappings but are not recognized by panphon library.

    You can list available mappings with --list-all or --list-ipa, or by visiting
    http://g2p-studio.herokuapp.com/api/v1/langs .
    """
    # Defer expensive imports
    from g2p.log import LOGGER
    from g2p.mappings import MAPPINGS_AVAILABLE
    from g2p.mappings.langs.utils import check_ipa_known_segs
    from g2p.mappings.utils import is_ipa

    if list_all or list_ipa:
        out_langs = sorted({x.out_lang for x in MAPPINGS_AVAILABLE})
        if list_ipa:
            out_langs = [x for x in out_langs if is_ipa(x)]
        LOGGER.info(
            "Specifying an output language will check all mappings into that language:\n"
        )
        for m in out_langs:
            print(f"{m}: ", end="")
            print(
                ("\n" + " " * len(m) + "  ").join(
                    sorted(x.in_lang for x in MAPPINGS_AVAILABLE if x.out_lang == m)
                )
            )
            print("")
        return

    for m in mapping:
        if m not in [x.out_lang for x in MAPPINGS_AVAILABLE]:
            raise click.UsageError(
                f"No known mappings into '{m}'. "
                "Use --list-all or --list-ipa to list valid options."
            )
        if not is_ipa(m):
            LOGGER.warning(
                f"No checks implemented yet for non-IPA mappings: '{m}' will not be checked."
            )

    if not mapping:
        LOGGER.info("Checking all IPA mappings.")
    else:
        LOGGER.info("Checking the following mappings: \n" + "\n".join(mapping))

    check_ipa_known_segs(list(mapping))


@click.option(
    "-i",
    "--in-dir",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    help=f"Scan DIRECTORY for mappings instead of the installed directory ({LANGS_DIR}).",
)
@click.option(
    "-o",
    "--out-dir",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    help=f"Output results in DIRECTORY instead of the installed directory ({LANGS_DIR}).",
)
@cli.command(context_settings=CONTEXT_SETTINGS)
def update(in_dir, out_dir):
    """Update cached language files."""
    # Defer expensive imports
    from g2p.mappings.langs import reload_db
    from g2p.mappings.langs.utils import cache_langs, network_to_echart

    if in_dir is None:
        in_dir = LANGS_DIR
    if out_dir is None:
        langs_path = os.path.join(in_dir, LANGS_FILE_NAME)
        network_path = os.path.join(in_dir, NETWORK_FILE_NAME)
    else:
        langs_path = os.path.join(out_dir, LANGS_FILE_NAME)
        network_path = os.path.join(out_dir, NETWORK_FILE_NAME)
    cache_langs(dir_path=in_dir, langs_path=langs_path, network_path=network_path)

    if in_dir == LANGS_DIR and out_dir is None:
        # We only update the documentation when updating using the default directories
        reload_db()
        network_to_echart(
            outfile=os.path.join(os.path.dirname(static_file), "languages-network.json")
        )  # updates g2p/static/languages-network.json


@click.option(
    "-o",
    "--out-dir",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    help=f"Output results in DIRECTORY instead of the installed directory ({Path(LANGS_DIR) / '..' / '.schema'}).",
)
@cli.command(context_settings=CONTEXT_SETTINGS)
def update_schema(out_dir):
    """Generate a schema for the model configuration.

    This should be done every time the model changes in a way that affects the schema.
    Changes to the schema should result in a minor version bump.
    But not every minor or major version bump requires a schema update.
    """
    # Defer expensive imports
    import pydantic

    from g2p.mappings import MappingConfig

    pydantic_major, pydantic_minor, _ = pydantic.VERSION.split(".", 3)
    if (int(pydantic_major), int(pydantic_minor)) >= (2, 9):
        print(f"Detected pydantic version {pydantic.VERSION} is greater than 2.8.")
        print("Please use Pydantic < 2.9 with Python < 3.14 to update schemas.")
        return

    # We shall not change the schema for patches, so only include major/minor version
    (major, minor, *_rest) = g2p._version.version_tuple
    major_minor = f"{major}.{minor}"

    # Determine path
    if out_dir is None:
        out_dir = Path(LANGS_DIR).parent / ".schema"
    else:
        out_dir = Path(out_dir)
    schema_path = out_dir / f"g2p-config-schema-{major_minor}.json"

    # Generate schema
    json_schema = MappingConfig.model_json_schema()
    # Add explicit schema dialect for SchemaStore that pydantic actually targets
    json_schema["$schema"] = "http://json-schema.org/draft-07/schema#"

    if schema_path.exists():
        with open(schema_path, encoding="utf8") as f:
            old_schema = json.load(f)
        if old_schema == json_schema:
            print(f"Schema {schema_path} is already up to date.")
        else:
            print(
                dedent(
                    f"""
                    Schema {schema_path}
                    exists for version {major_minor} but is not up to date.  If it was already published to
                    the SchemaStore (see
                    https://github.com/SchemaStore/schemastore/blob/master/src/api/json/catalog.json)
                    then you must bump the minor or major version number of g2p and generate the
                    schema again.  If not and you want to overwrite it, please delete the existing
                    schema and try again.
                    """
                ),
                file=sys.stderr,
            )
            raise click.UsageError("Schema already exists but is not up to date.")
    else:
        prev_schema_files = sorted(out_dir.glob("g2p-config-schema-*.json"))
        if prev_schema_files:
            with open(prev_schema_files[-1], encoding="utf8") as f:
                old_schema = json.load(f)
        else:
            old_schema = None
        if old_schema == json_schema:
            print(f"Schema {prev_schema_files[-1]}")
            print(
                f"is still up to date.  No need to generate a new schema for version {major_minor}."
            )
        else:
            with open(schema_path, "w", encoding="ascii") as f:
                json.dump(json_schema, f, indent=2)
            print(f"Wrote {schema_path}.")


@click.argument("path", type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.argument("lang")
@cli.command(
    context_settings=CONTEXT_SETTINGS,
    short_help="Scan a document for unknown characters.",
)
def scan(lang, path) -> None:
    """Scan a document for non target language characters.

    Displays the set of un-mapped characters in a document.
    Accounts for case sensitivity in the configuration.
    """
    # Defer expensive imports
    from g2p.log import LOGGER
    from g2p.mappings import MAPPINGS_AVAILABLE, Mapping
    from g2p.mappings.langs import LANGS_NETWORK
    from g2p.mappings.utils import normalize

    # Check input lang exists
    if lang not in LANGS_NETWORK.nodes:
        raise click.UsageError(f"'{lang}' is not a valid value for 'LANG'")

    # Retrieve the mappings for lang
    case_sensitive = True
    mappings: List[Mapping] = []
    for mapping in MAPPINGS_AVAILABLE:
        mapping_name = mapping.in_lang
        # Exclude mappings for converting between IPAs
        if mapping_name.startswith(lang) and "ipa" not in mapping_name:
            case_sensitive = case_sensitive and mapping.case_sensitive
            mappings.append(mapping)

    # Get input chars in mapping
    mapped_chars = set()
    for lang_mapping in mappings:
        for x in lang_mapping.rules:
            mapped_chars.add(normalize(x.rule_input, "NFD"))
    # Find unmapped chars
    filter_chars = " \n"
    mapped_string = "".join(mapped_chars)
    pattern = f"[^{mapped_string}{filter_chars}.]"
    prog = re.compile(pattern)

    with open(path, "r", encoding="utf8") as file:
        data = normalize(file.read(), "NFD")
        if not case_sensitive:
            data = data.lower()
        unmapped = set(prog.findall(data))
        if unmapped:
            LOGGER.warning("The following characters are not mapped:")
            print(unmapped)


@click.option("--csv", is_flag=True, help="Output mappings in CSV format.")
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="Display mapping configs and all rules too.",
)
@click.argument("lang2", required=False, default=None)
@click.argument("lang1", required=False, default=None)
@cli.command(context_settings=CONTEXT_SETTINGS, short_help="Show cached mappings.")
def show_mappings(lang1, lang2, verbose, csv):
    """Show cached mappings, as last updated by "g2p update".

    Mappings on the path from LANG1 to LANG2 are displayed.
    If only LANG1 is used, all mappings to or from LANG1 are displayed.
    With no LANG, all cached mappings are included.
    """
    # Defer expensive imports
    from g2p.mappings import MAPPINGS_AVAILABLE, Mapping
    from g2p.transducer import CompositeTransducer, Transducer

    if lang1 is not None and lang2 is not None:
        try:
            transducer = make_g2p(lang1, lang2, tokenize=False)
        except (NoPath, InvalidLanguageCode) as e:
            raise click.UsageError(
                f'Cannot find mapping from "{lang1}" to "{lang2}": {e}'
            ) from e

        if isinstance(transducer, Transducer):
            mappings = [transducer.mapping]
        else:
            assert isinstance(transducer, CompositeTransducer)
            mappings = [t.mapping for t in transducer._transducers]

    elif lang1 is not None:
        mappings = [
            Mapping.find_mapping(in_lang=m.in_lang, out_lang=m.out_lang)
            for m in MAPPINGS_AVAILABLE
            if m.in_lang == lang1 or m.out_lang == lang1
        ]
        if not mappings:
            raise click.BadParameter(
                f'No mapping found to or from "{lang1}".', param_hint="lang1"
            )

    else:
        mappings = [
            Mapping.find_mapping(in_lang=m.in_lang, out_lang=m.out_lang)
            for m in MAPPINGS_AVAILABLE
        ]

    file_type = "csv" if csv else "json"
    if verbose:
        for m in mappings:
            json.dump(
                json.loads(m.model_dump_json(exclude={"rules": True})),
                sys.stdout,
                indent=4,
                ensure_ascii=False,
            )
            print()
            m.mapping_to_stream(sys.stdout, file_type=file_type)
            print()
            print()
    else:
        for i, m in enumerate(mappings):
            print(f"{i+1}: {m.in_lang} → {m.out_lang}  ({m.display_name})")



#!/usr/bin/env python

from unittest import TestCase, main

import g2p


class TokenizeAndMapTest(TestCase):
    """Test suite for chaining tokenization and transduction"""

    def setUp(self):
        pass

    def contextualize(self, word: str):
        return word + " " + word + " ," + word + ", " + word

    def test_tok_and_map_fra(self):
        """Chaining tests: tokenize and map a string"""
        transducer = g2p.make_g2p("fra", "fra-ipa", tokenize=False)
        tokenizer = g2p.make_tokenizer("fra")
        # "teste" in isolation is at string and word end and beginning
        word_ipa = transducer("teste").output_string
        # "teste" followed by space or punctuation should be mapped to the same string
        string_ipa = g2p.tokenize_and_map(
            tokenizer, transducer, self.contextualize("teste")
        )
        self.assertEqual(string_ipa, self.contextualize(word_ipa))

    def test_tok_and_map_mic(self):
        transducer = g2p.make_g2p("mic", "mic-ipa", tokenize=False)
        tokenizer = g2p.make_tokenizer("mic")
        word_ipa = transducer("sq").output_string
        string_ipa = g2p.tokenize_and_map(
            tokenizer, transducer, self.contextualize("sq")
        )
        self.assertEqual(string_ipa, self.contextualize(word_ipa))

    def test_tokenizing_transducer(self) -> None:
        from g2p.transducer import TokenizingTransducer

        ref_word_ipa = g2p.make_g2p("mic", "mic-ipa", tokenize=False)(
            "sq"
        ).output_string
        transducer = g2p.make_g2p("mic", "mic-ipa")  # tokenizes on "mic" via "path"
        assert isinstance(transducer, TokenizingTransducer)
        self.assertEqual(transducer.transducer.in_lang, transducer.in_lang)
        self.assertEqual(transducer.transducer.out_lang, transducer.out_lang)
        self.assertEqual(transducer.transducer, transducer.transducers[0])
        word_ipa = transducer("sq").output_string
        self.assertEqual(word_ipa, ref_word_ipa)
        string_ipa = transducer(self.contextualize("sq")).output_string
        self.assertEqual(string_ipa, self.contextualize(ref_word_ipa))

    def test_tokenizing_transducer_chain(self):
        transducer = g2p.make_g2p("fra", "eng-arpabet")
        self.assertEqual(
            self.contextualize(transducer("teste").output_string),
            transducer(self.contextualize("teste")).output_string,
        )

    def test_tokenizing_transducer_debugger(self):
        transducer = g2p.make_g2p("fra", "fra-ipa")
        debugger = transducer("ceci est un test.").debugger
        self.assertEqual(len(debugger), 4)

    def test_tokenizing_transducer_edges(self):
        transducer = g2p.make_g2p("fra", "fra-ipa")
        tg = transducer("est est")
        # est -> ɛ, so edges are (0, 0), (1, 0), (2, 0) for each "est", plus the
        # space to the space, and the second set of edges being offset
        ref_edges = [(0, 0), (1, 0), (2, 0), (3, 1), (4, 2), (5, 2), (6, 2)]
        self.assertEqual(tg.edges, ref_edges)
        ref_alignments = [("est", "ɛ"), (" ", " "), ("est", "ɛ")]
        self.assertEqual(tg.substring_alignments(), ref_alignments)

    def test_tokenizing_transducer_edges2(self):
        ref_edges = g2p.make_g2p("fra", "fra-ipa", tokenize=False)("ça ça").edges
        edges = g2p.make_g2p("fra", "fra-ipa")("ça ça").edges
        self.assertEqual(edges, ref_edges)

    def test_tokenizing_transducer_edge_chain(self):
        transducer = g2p.make_g2p("fra", "eng-arpabet")
        # .edges on a transducer is always a single array with the
        # end-to-end mapping, for a composed transducer we can access
        # the individual tiers with .tiers
        ref_edges = [
            # "est est" -> "EH  EH "
            # est -> EH
            (0, 0),
            (0, 1),
            (0, 2),
            (1, 0),
            (1, 1),
            (1, 2),
            (2, 0),
            (2, 1),
            (2, 2),
            # space -> space
            (3, 3),
            # est -> EH
            (4, 4),
            (4, 5),
            (4, 6),
            (5, 4),
            (5, 5),
            (5, 6),
            (6, 4),
            (6, 5),
            (6, 6),
        ]
        tg = transducer("est est")
        self.assertEqual(tg.alignments(), ref_edges)
        ref_alignments = [("est", "EH "), (" ", " "), ("est", "EH ")]
        self.assertEqual(tg.substring_alignments(), ref_alignments)

        tier_edges = [x.edges for x in tg.tiers]
        ref_tier_edges = [
            # "est est" -> "ɛ ɛ"
            [(0, 0), (1, 0), (2, 0), (3, 1), (4, 2), (5, 2), (6, 2)],
            # "ɛ ɛ" -> "ɛ ɛ"
            [(0, 0), (1, 1), (2, 2)],
            # "ɛ ɛ" -> "EH  EH "
            [(0, 0), (0, 1), (0, 2), (1, 3), (2, 4), (2, 5), (2, 6)],
        ]
        self.assertEqual(tier_edges, ref_tier_edges)
        tier_alignments = [x.substring_alignments() for x in tg.tiers]
        ref_tier_alignments = [
            [("est", "ɛ"), (" ", " "), ("est", "ɛ")],
            [("ɛ", "ɛ"), (" ", " "), ("ɛ", "ɛ")],
            [("ɛ", "EH "), (" ", " "), ("ɛ", "EH ")],
        ]
        self.assertEqual(tier_alignments, ref_tier_alignments)

    def test_tokenizing_transducer_edge_spaces(self):
        transducer = g2p.make_g2p("fra", "eng-arpabet")
        ref_edges = [
            # "  a, " -> "  AA , "
            (0, 0),
            (1, 1),
            (2, 2),
            (2, 3),
            (2, 4),
            (3, 5),
            (4, 6),
        ]
        self.assertEqual(transducer("  a, ").alignments(), ref_edges)
        tier_edges = [x.edges for x in transducer("  a, ").tiers]
        ref_tier_edges = [
            # "  a, " -> "  a, "
            [(0, 0), (1, 1), (2, 2), (3, 3), (4, 4)],
            # "  a, " -> "  ɑ, "
            [(0, 0), (1, 1), (2, 2), (3, 3), (4, 4)],
            # "  ɑ, " -> "  AA , "
            [(0, 0), (1, 1), (2, 2), (2, 3), (2, 4), (3, 5), (4, 6)],
        ]
        self.assertEqual(tier_edges, ref_tier_edges)

    def test_removed_tok_langs_in_v2(self) -> None:
        # monkey patch to make sure we always exercise the TypeError pathway
        saved_version = g2p._version.VERSION
        g2p._version.VERSION = "2.0"
        with self.assertRaises(TypeError):
            _ = g2p.make_g2p("iku-sro", "eng-ipa", "path")  # type: ignore
        g2p._version.VERSION = saved_version

    def test_make_g2p_cache(self):
        self.assertIs(
            g2p.make_g2p("fra", "fra-ipa", tokenize=False),
            g2p.make_g2p("fra", "fra-ipa", tokenize=False),
        )

        self.assertIsNot(
            g2p.make_g2p("fra", "fra-ipa", tokenize=True),
            g2p.make_g2p("fra", "fra-ipa", tokenize=False),
        )
        my_tokenizer = g2p.make_tokenizer("oji")
        self.assertIs(
            g2p.make_g2p("oji", "eng-ipa", custom_tokenizer=my_tokenizer),
            g2p.make_g2p("oji", "eng-ipa", custom_tokenizer=my_tokenizer),
        )
        self.assertIs(
            g2p.make_g2p("oji", "eng-ipa", custom_tokenizer=my_tokenizer),
            g2p.make_g2p("oji", "eng-ipa", custom_tokenizer=g2p.make_tokenizer("oji")),
        )
        self.assertIsNot(
            g2p.make_g2p("oji", "eng-ipa", custom_tokenizer=my_tokenizer),
            g2p.make_g2p("oji", "eng-ipa", custom_tokenizer=g2p.make_tokenizer("ikt")),
        )
        self.assertIsNot(
            g2p.make_g2p("oji", "eng-ipa", custom_tokenizer=my_tokenizer),
            g2p.make_g2p("oji", "eng-ipa", tokenize=True),
        )


if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""Organize tests into Test Suites

Run with "python run.py <suite>" where <suite> can be all, dev, or a few other
options (see run_tests() for the full list).

Add --describe to list the contents of the selected suite instead of running it.
"""

import argparse
import os
import re
import sys
from unittest import TestLoader, TestSuite, TextTestRunner

# Unit tests
from g2p.log import LOGGER
from g2p.tests.test_check_ipa_arpabet import CheckIpaArpabetTest
from g2p.tests.test_cli import CliTest
from g2p.tests.test_create_mapping import MappingCreationTest
from g2p.tests.test_doctor import DoctorTest
from g2p.tests.test_doctor_expensive import ExpensiveDoctorTest
from g2p.tests.test_fallback import FallbackTest
from g2p.tests.test_indices import IndicesTest
from g2p.tests.test_langs import LangTest
from g2p.tests.test_lexicon_transducer import LexiconTransducerTest
from g2p.tests.test_mappings import MappingTest
from g2p.tests.test_network import NetworkLiteTest, NetworkTest
from g2p.tests.test_neural import NeuralLangTest
from g2p.tests.test_tokenize_and_map import TokenizeAndMapTest
from g2p.tests.test_tokenizer import TokenizerTest
from g2p.tests.test_transducer import TransducerTest
from g2p.tests.test_unidecode_transducer import UnidecodeTransducerTest
from g2p.tests.test_utils import UtilsTest
from g2p.tests.test_z_local_config import LocalConfigTest

if sys.version_info >= (3, 8, 0):
    from g2p.tests.test_api_resources import ResourceIntegrationTest
    from g2p.tests.test_api_v2 import TestAPIV2

    API_TEST_CLASSES = [ResourceIntegrationTest, TestAPIV2]
else:
    API_TEST_CLASSES = []

LOADER = TestLoader()

TRANSDUCER_TESTS = [
    LOADER.loadTestsFromTestCase(test)
    for test in [
        IndicesTest,
        TransducerTest,
        UnidecodeTransducerTest,
        LexiconTransducerTest,
    ]
]

MAPPINGS_TESTS = [
    LOADER.loadTestsFromTestCase(test)
    for test in [
        FallbackTest,
        MappingCreationTest,
        MappingTest,
        NetworkTest,
        NetworkLiteTest,
        UtilsTest,
        TokenizerTest,
        TokenizeAndMapTest,
        CheckIpaArpabetTest,
    ]
]

LANGS_TESTS = [
    LOADER.loadTestsFromTestCase(test)
    for test in [
        LangTest,
    ]
]

# We are excluding these tests from the dev suite because
# they require torch and other heavy dependencies
# and the tests also require downloading large g2p models
NEURAL_TESTS = [LOADER.loadTestsFromTestCase(test) for test in [NeuralLangTest]]

INTEGRATION_TESTS = [
    LOADER.loadTestsFromTestCase(test)
    for test in [
        CliTest,
        DoctorTest,
        ExpensiveDoctorTest,
    ]
    + API_TEST_CLASSES
]

# LocalConfigTest has to get run last, to avoid interactions with other test
# cases, since it has side effects on the global database
LAST_DEV_TEST = [
    LOADER.loadTestsFromTestCase(test)
    for test in [
        LocalConfigTest,
    ]
]

DEV_TESTS = (
    TRANSDUCER_TESTS + MAPPINGS_TESTS + LANGS_TESTS + INTEGRATION_TESTS + LAST_DEV_TEST
)


def list_tests(suite: TestSuite):
    for subsuite in suite:
        for match in re.finditer(r"tests=\[([^][]+)\]>", str(subsuite)):
            for test_case in match[1].split(", "):
                yield test_case.replace("g2p.tests.", "")


def describe_suite(suite: TestSuite):
    full_suite = LOADER.discover(os.path.dirname(__file__))
    full_list = list(list_tests(full_suite))
    requested_list = list(list_tests(suite))
    requested_set = set(requested_list)
    print("Test suite includes:", *sorted(requested_list), sep="\n")
    print(
        "\nTest suite excludes:",
        *sorted(test for test in full_list if test not in requested_set),
        sep="\n",
    )


SUITES = ("all", "dev", "integ", "langs", "mappings", "trans", "neural")


def run_tests(suite: str, describe: bool = False, verbosity: int = 3) -> bool:
    """Run the test suite specified in suite.

    Args:
        suite: one of SUITES, "dev" if the empty string
        describe: if True, list all the test cases instead of running them.

    Returns: Bool: True iff success
    """
    if not suite:
        LOGGER.info("No test suite specified, defaulting to dev.")
        suite = "dev"

    if suite == "all":
        test_suite = LOADER.discover(os.path.dirname(__file__))
    elif suite == "trans":
        test_suite = TestSuite(TRANSDUCER_TESTS)
    elif suite == "langs":
        test_suite = TestSuite(LANGS_TESTS)
    elif suite == "mappings":
        test_suite = TestSuite(MAPPINGS_TESTS)
    elif suite == "integ":
        test_suite = TestSuite(INTEGRATION_TESTS)
    elif suite == "neural":
        test_suite = TestSuite(NEURAL_TESTS)
    elif suite == "dev":
        test_suite = TestSuite(DEV_TESTS)
    else:
        LOGGER.error("Please specify a test suite to run among: " + ", ".join(SUITES))
        return False

    if describe:
        describe_suite(test_suite)
        return True
    else:
        runner = TextTestRunner(verbosity=verbosity)
        success = runner.run(test_suite).wasSuccessful()
        if not success:
            LOGGER.error("Some tests failed. Please see log above.")
        return success


def main():
    parser = argparse.ArgumentParser(description="Run g2p test suites.")
    parser.add_argument("--quiet", "-q", action="store_true", help="reduce output")
    parser.add_argument(
        "--describe", action="store_true", help="describe the selected test suite"
    )
    parser.add_argument(
        "suite",
        nargs="?",
        default="dev",
        help="the test suite to run [dev]",
        choices=SUITES,
    )
    args = parser.parse_args()
    result = run_tests(args.suite, args.describe, 1 if args.quiet else 3)
    if not result:
        sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

from unittest import TestCase, main

from g2p.mappings import Mapping, Rule
from g2p.mappings.create_fallback_mapping import align_to_dummy_fallback


class FallbackTest(TestCase):
    """Basic Mapping Fallback Test.
    This feature is experimental, but it will try to map each character
    in a Mapping to one of the following 'unmarked' phonemes:
        ["ɑ", "i", "u", "t", "s", "n"]
    If the mapping is 'ipa', it will align the inventories directly.
    If not, it will take a best guess at what the Unicode character using Unidecode and then align from there.
    """

    def setUp(self):
        pass

    def test_mapping(self):
        self.maxDiff = None
        mapping = Mapping(
            rules=[
                {"in": "a", "out": "æ"},
                {"in": "e", "out": "ɐ"},
                {"in": "i", "out": "ɑ̃"},
                {"in": "b", "out": "β"},
                {"in": "g", "out": "ɡ"},
                {"in": "g", "out": "g"},
                {"in": "i", "out": "ةُ"},
            ],
            in_lang="test",
            out_lang="test-out",
        )
        ipa_mapping = Mapping(
            rules=[
                {"in": "a", "out": "æ"},
                {"in": "e", "out": "ɐ"},
                {"in": "i", "out": "ɑ̃"},
                {"in": "b", "out": "β"},
                {"in": "g", "out": "ɡ"},
            ],
            in_lang="test",
            out_lang="test-ipa",
        )
        test_in = align_to_dummy_fallback(mapping, quiet=True)
        self.assertEqual(
            test_in.rules,
            [
                Rule(rule_input="a", rule_output="ɑ", match_pattern="a"),
                Rule(rule_input="e", rule_output="i", match_pattern="e"),
                Rule(rule_input="i", rule_output="i", match_pattern="i"),
                Rule(rule_input="b", rule_output="t", match_pattern="b"),
                Rule(rule_input="g", rule_output="t", match_pattern="g"),
                Rule(rule_input="g", rule_output="t", match_pattern="g"),
                Rule(rule_input="i", rule_output="i", match_pattern="i"),
            ],
        )

        test_out = align_to_dummy_fallback(mapping, "out", quiet=True)
        self.assertEqual(
            test_out.rules,
            [
                Rule(rule_input="æ", rule_output="ɑi", match_pattern="æ"),
                Rule(rule_input="ɐ", rule_output="ɑ", match_pattern="ɐ"),
                Rule(rule_input="ɑ̃", rule_output="ɑ", match_pattern="ɑ̃"),
                Rule(rule_input="β", rule_output="t", match_pattern="β"),
                Rule(rule_input="ɡ", rule_output="t", match_pattern="ɡ"),
                Rule(rule_input="g", rule_output="t", match_pattern="g"),
                Rule(rule_input="ةُ", rule_output="ɑu", match_pattern="ةُ"),
            ],
        )
        test_ipa = align_to_dummy_fallback(ipa_mapping, "out", quiet=True)
        panphon_021_ref = [
            Rule(rule_input="æ", rule_output="ɑ", match_pattern="æ"),
            Rule(rule_input="ɐ", rule_output="i", match_pattern="ɐ"),
            Rule(rule_input="ɑ̃", rule_output="ɑ", match_pattern="ɑ̃"),
            Rule(rule_input="β", rule_output="s", match_pattern="β"),
            Rule(rule_input="ɡ", rule_output="t", match_pattern="ɡ"),
        ]
        panphon_020_ref = [
            panphon_021_ref[0],
            Rule(rule_input="ɐ", rule_output="ɑ", match_pattern="ɐ"),
            *panphon_021_ref[2:],
        ]
        if test_ipa.rules != panphon_021_ref:
            self.assertEqual(test_ipa.rules, panphon_020_ref)


if __name__ == "__main__":
    main()
